import re
from gi.repository import Gtk, GObject, Pango
from pychess.widgets import insert_formatted

TYPE_PERSONAL, TYPE_CHANNEL, TYPE_GUEST, TYPE_ADMIN, TYPE_COMP, TYPE_BLINDFOLD = range(
    6
)


def get_playername(playername):
    re_m = re.match(r"(\w+)\W*", playername)
    return re_m.groups()[0]


class BulletCellRenderer(Gtk.CellRenderer):
    __gproperties__ = {
        "color": (
            object,
            "Color",
            "Color",
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.color = None
        self.width = 16
        self.height = 16

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def do_render(self, context, widget, bg_area, cell_area, flags):
        if self.color is None:
            return
        else:
            red, green, blue = self.color

        x_loc, y_loc = self.get_size(widget, cell_area)[:2]

        context.set_source_rgb(red, green, blue)
        context.rectangle(x_loc, y_loc, self.width, self.height)
        context.fill()

        context.set_line_width(1)
        context.set_source_rgba(0, 0, 0, 0.5)
        context.rectangle(x_loc + 0.5, y_loc + 0.5, self.width - 1, self.height - 1)
        context.stroke()

        context.set_line_width(1)
        context.set_source_rgba(1, 1, 1, 0.5)
        context.rectangle(x_loc + 1.5, y_loc + 1.5, self.width - 3, self.height - 3)
        context.stroke()

    def on_get_size(self, widget, cell_area=None):
        if cell_area:
            y_loc = int(cell_area.height / 2.0 - self.height / 2.0) + cell_area.y
            x_loc = cell_area.x
        else:
            y_loc = 0
            x_loc = 0
        return (x_loc + 1, y_loc + 1, self.width + 2, self.height + 2)


GObject.type_register(BulletCellRenderer)


class Panel:
    def start(self):
        pass

    def addItem(self, id, text, type, chat_view):
        pass

    def removeItem(self, id):
        pass

    def selectItem(self, id):
        pass


class InfoPanel(Gtk.Notebook, Panel):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.set_show_tabs(False)
        self.set_show_border(False)
        self.id2Widget = {}

        label = Gtk.Label()
        label.set_markup("<big>%s</big>" % _("No conversation's selected"))
        label.props.xalign = 0.5
        label.props.yalign = 0.381966011
        label.props.justify = Gtk.Justification.CENTER
        label.props.wrap = True
        label.props.width_request = 115
        self.append_page(label, None)

        self.connection = connection

    def addItem(self, grp_id, text, grp_type, chat_view):
        if grp_type in (
            TYPE_PERSONAL,
            TYPE_COMP,
            TYPE_GUEST,
            TYPE_ADMIN,
            TYPE_BLINDFOLD,
        ):
            infoItem = self.PlayerInfoItem(grp_id, text, chat_view, self.connection)
        elif grp_type == TYPE_CHANNEL:
            infoItem = self.ChannelInfoItem(grp_id, text, chat_view, self.connection)
        self.addPage(infoItem, grp_id)

    def removeItem(self, grp_id):
        self.removePage(grp_id)

    def selectItem(self, grp_id):
        child = self.id2Widget.get(grp_id)
        if child is not None:
            self.set_current_page(self.page_num(child))

    def addPage(self, widget, grp_id):
        self.id2Widget[grp_id] = widget
        self.append_page(widget, None)
        widget.show_all()

    def removePage(self, grp_id):
        child = self.id2Widget.pop(grp_id)
        self.remove_page(self.page_num(child))

    class PlayerInfoItem(Gtk.Alignment):
        def __init__(self, id, text, chat_view, connection):
            GObject.GObject.__init__(self, xscale=1, yscale=1)
            self.add(Gtk.Label(label=_("Loading player data")))

            playername = get_playername(text)
            self.fm = connection.fm
            self.handle_id = self.fm.connect(
                "fingeringFinished", self.onFingeringFinished, playername
            )

            self.fm.finger(playername)

        def onFingeringFinished(self, fm, finger, playername):
            if (
                not isinstance(self.get_child(), Gtk.Label)
                or finger.getName().lower() != playername.lower()
            ):
                return
            self.fm.disconnect(self.handle_id)

            label = Gtk.Label()
            label.set_markup("<b>%s</b>" % playername)
            widget = Gtk.Frame()
            widget.set_label_widget(label)
            widget.set_shadow_type(Gtk.ShadowType.NONE)

            alignment = Gtk.Alignment.new(0, 0, 1, 1)
            alignment.set_padding(3, 0, 12, 0)
            widget.add(alignment)

            text_view = Gtk.TextView()
            text_view.set_editable(False)
            text_view.set_cursor_visible(False)
            text_view.props.wrap_mode = Gtk.WrapMode.WORD

            tb_iter = text_view.get_buffer().get_end_iter()
            for i, note in enumerate(finger.getNotes()):
                if note:
                    insert_formatted(text_view, tb_iter, "%s\n" % note)
            text_view.show_all()
            scroll_win = Gtk.ScrolledWindow()
            scroll_win.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            scroll_win.add(text_view)
            alignment.add(scroll_win)

            self.remove(self.get_child())
            self.add(widget)
            widget.show_all()

    class ChannelInfoItem(Gtk.Alignment):
        def __init__(self, grp_id, text, chat_view, connection):
            GObject.GObject.__init__(self, xscale=1, yscale=1)
            self.cm = connection.cm
            self.add(Gtk.Label(label=_("Receiving list of players")))

            self.names = set()
            chat_view.connect("messageAdded", self.onMessageAdded)
            self.store = Gtk.ListStore(
                object, str, bool  # (r,g,b) Color tuple  # name string  # is separator
            )

            connection.players.connect("FICSPlayerExited", self.onPlayerRemoved)

            self.handle_id = self.cm.connect(
                "receivedNames", self.onNamesReceived, grp_id
            )
            self.cm.getPeopleInChannel(grp_id)

        def onPlayerRemoved(self, players, player):
            if player.name in self.names:
                for row in self.store:
                    if row[1] == player.name:
                        self.store.remove(row.iter)
                        break
                self.names.remove(player.name)

        def onNamesReceived(self, cm, channel, people, channel_):
            if not isinstance(self.get_child(), Gtk.Label) or channel != channel_:
                return
            cm.disconnect(self.handle_id)

            scroll_win = Gtk.ScrolledWindow()
            scroll_win.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

            tv_list = Gtk.TreeView()
            tv_list.set_headers_visible(False)
            tv_list.set_tooltip_column(1)
            tv_list.set_model(self.store)
            tv_list.append_column(Gtk.TreeViewColumn("", BulletCellRenderer(), color=0))
            cell = Gtk.CellRendererText()
            cell.props.ellipsize = Pango.EllipsizeMode.END
            tv_list.append_column(Gtk.TreeViewColumn("", cell, text=1))
            tv_list.fixed_height_mode = True

            self.separatorIter = self.store.append([(), "", True])

            tv_list.set_row_separator_func(lambda m, i, d: m.get_value(i, 2), None)
            scroll_win.add(tv_list)

            self.store.connect("row-inserted", lambda w, p, i: tv_list.queue_resize())
            self.store.connect("row-deleted", lambda w, i: tv_list.queue_resize())

            # Add those names. If this is not the first namesReceive, we only
            # add the new names
            noneed = {name for (color, name, isSeparator) in self.store}
            for name in people:
                if name in noneed:
                    continue
                self.store.append([(1, 1, 1), name, False])
                self.names.add(name)

            self.remove(self.get_child())
            self.add(scroll_win)
            self.show_all()

        def onMessageAdded(self, chat_view, sender, text, color):
            s_iter = self.store.get_iter_first()

            # If the names list hasn't been retrieved yet, we have to skip this
            if not s_iter:
                return

            while self.store.get_path(s_iter) != self.store.get_path(
                self.separatorIter
            ):
                person = self.store.get_value(s_iter, 1)
                # If the person is already in the area before the separator, we
                # don't have to do anything
                if person.lower() == sender.lower():
                    return
                s_iter = self.store.iter_next(s_iter)

            # Go to s_iter after separator
            s_iter = self.store.iter_next(s_iter)

            while s_iter and self.store.iter_is_valid(s_iter):
                person = self.store.get_value(s_iter, 1)
                if person.lower() == sender.lower():
                    self.store.set_value(s_iter, 0, color)
                    self.store.move_before(s_iter, self.separatorIter)
                    return
                s_iter = self.store.iter_next(s_iter)

            # If the person was not in the area under the separator of the
            # store, it must be a new person, who has joined the channel, and we
            # simply add him before the separator
            self.store.insert_before(self.separatorIter, [color, sender, False])



from pychess.Utils.const import PLACEMENTCHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board


# Placement chess (Bronstein/Benko/Pre-chess)
# http://www.quantumgambitz.com/blog/chess/cga/bronstein-chess-pre-chess-shuffle-chess
PLACEMENTSTART = "8/pppppppp/8/8/8/8/PPPPPPPP/8/nnbbrrqkNNBBRRQK w - - 0 1"


class PlacementBoard(Board):
    variant = PLACEMENTCHESS
    __desc__ = _(
        "Pre-chess: https://en.wikipedia.org/wiki/List_of_chess_variants#Different_starting_position"
    )
    name = _("Placement")
    cecp_name = "placement"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=PLACEMENTSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



# Crazyhouse Chess

from pychess.Utils.const import CRAZYHOUSECHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board


class CrazyhouseBoard(Board):
    variant = CRAZYHOUSECHESS
    __desc__ = _(
        "FICS crazyhouse: http://www.freechess.org/Help/HelpFiles/crazyhouse.html"
    )
    name = _("Crazyhouse")
    cecp_name = "crazyhouse"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD



from gi.repository import Gtk, GLib, GdkPixbuf

from pychess.ic.FICSObjects import FICSGame, get_player_tooltip_text
from pychess.ic import (
    TYPE_BLITZ,
    TYPE_LIGHTNING,
    TYPE_STANDARD,
    RATING_TYPES,
    TYPE_BULLET,
    TYPE_ONE_MINUTE,
    TYPE_THREE_MINUTE,
    TYPE_FIVE_MINUTE,
    TYPE_FIFTEEN_MINUTE,
    TYPE_FORTYFIVE_MINUTE,
)
from pychess.Utils.IconLoader import load_icon, get_pixbuf
from pychess.System.Log import log
from pychess.System import conf, uistuff
from pychess.System.prefix import addDataPrefix
from pychess.perspectives.fics.ParrentListSection import (
    ParrentListSection,
    cmp,
    SEPARATOR,
    FOLLOW,
    FINGER,
    ARCHIVED,
    OBSERVE,
)


__title__ = _("Game List")

__icon__ = addDataPrefix("glade/panel_games.svg")

__desc__ = _("List of ongoing games")


class Sidepanel(ParrentListSection):
    def load(self, widgets, connection, lounge):
        self.widgets = widgets
        self.connection = connection
        self.lounge = lounge

        __widget__ = lounge.games_list

        self.games = {}
        self.recpix = load_icon(16, "media-record")
        self.clearpix = get_pixbuf("glade/board.png")
        self.tv = self.widgets["gametreeview"]
        self.store = Gtk.ListStore(
            FICSGame, GdkPixbuf.Pixbuf, str, int, str, int, str, str
        )

        self.game_filter = self.store.filter_new()
        self.game_filter.set_visible_func(self.game_filter_func)

        self.filter_toggles = {}
        self.filter_buttons = (
            "standard_toggle",
            "blitz_toggle",
            "lightning_toggle",
            "variant_toggle",
        )
        for widget in self.filter_buttons:
            uistuff.keep(self.widgets[widget], widget)
            self.widgets[widget].connect("toggled", self.on_filter_button_toggled)
            initial = conf.get(widget)
            self.filter_toggles[widget] = initial
            self.widgets[widget].set_active(initial)

        self.model = self.game_filter.sort_new_with_model()
        self.tv.set_model(self.model)
        self.tv.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        self.addColumns(
            self.tv,
            "FICSGame",
            "",
            _("White"),
            _("Rating"),
            _("Black"),
            _("Rating"),
            _("Type"),
            _("Rated"),
            hide=[0],
            pix=[1],
        )

        self.tv.get_model().set_sort_func(0, self.pixCompareFunction, 1)
        for i in range(1, 7):
            self.tv.get_model().set_sort_func(i, self.compareFunction, i)
        self.prev_sort_column_id = []
        self.model.connect("sort-column-changed", self.on_sort_column_change)

        self.tv.set_has_tooltip(True)
        self.tv.connect("query-tooltip", self.on_query_tooltip)
        self.selection = self.tv.get_selection()
        self.selection.connect("changed", self.onSelectionChanged)
        self.onSelectionChanged(self.selection)

        try:
            self.tv.set_search_position_func(self.lowLeftSearchPosFunc, None)
        except AttributeError:
            # Unknow signal name is raised by gtk < 2.10
            pass

        def searchCallback(model, column, key, sel_iter, user_data):
            if model.get_value(sel_iter, 2).lower().startswith(key) or model.get_value(
                sel_iter, 4
            ).lower().startswith(key):
                return False
            return True

        self.tv.set_search_equal_func(searchCallback, None)

        self.connection.games.connect("FICSGameCreated", self.onGameAdd)
        self.connection.games.connect("FICSGameEnded", self.onGameRemove)
        self.widgets["observeButton"].connect("clicked", self.on_observe)
        self.tv.connect("row-activated", self.on_observe)
        self.connection.bm.connect("obsGameCreated", self.onGameObserved)
        self.connection.bm.connect("obsGameUnobserved", self.onGameUnobserved)

        self.tv.connect("button-press-event", self.button_press_event)
        self.createLocalMenu((OBSERVE, FOLLOW, SEPARATOR, FINGER, ARCHIVED))

        return __widget__

    def game_filter_func(self, model, iter, data):
        game = model[iter][0]
        is_standard = game.game_type.rating_type in (
            TYPE_STANDARD,
            TYPE_FIFTEEN_MINUTE,
            TYPE_FORTYFIVE_MINUTE,
        )
        is_blitz = game.game_type.rating_type in (
            TYPE_BLITZ,
            TYPE_THREE_MINUTE,
            TYPE_FIVE_MINUTE,
        )
        is_lightning = game.game_type.rating_type in (
            TYPE_LIGHTNING,
            TYPE_BULLET,
            TYPE_ONE_MINUTE,
        )
        is_variant = game.game_type.rating_type in RATING_TYPES[9:]
        return (
            (self.filter_toggles["standard_toggle"] and is_standard)
            or (self.filter_toggles["blitz_toggle"] and is_blitz)
            or (self.filter_toggles["lightning_toggle"] and is_lightning)
            or (self.filter_toggles["variant_toggle"] and is_variant)
        )

    def on_filter_button_toggled(self, widget):
        for button in self.filter_buttons:
            self.filter_toggles[button] = self.widgets[button].get_active()
        self.game_filter.refilter()

    # Multi-column sort based on TreeModelSortUtil from
    # https://github.com/metomi/rose/blob/master/lib/python/rose/gtk/util.py
    def on_sort_column_change(self, model):
        """Store previous sorting information for multi-column sorts."""
        id, order = self.tv.get_model().get_sort_column_id()
        if id is None and order is None:
            return False
        if self.prev_sort_column_id and self.prev_sort_column_id[0][0] == id:
            self.prev_sort_column_id.pop(0)
        self.prev_sort_column_id.insert(0, (id, order))
        if len(self.prev_sort_column_id) > 2:
            self.prev_sort_column_id.pop()

    def compareFunction(self, model, iter0, iter1, column):
        """Multi-column sort."""
        val0 = model.get_value(iter0, column)
        val1 = model.get_value(iter1, column)
        rval = cmp(val0, val1)
        # If rval is 1 or -1, no need for a multi-column sort.
        if rval == 0:
            this_order = self.tv.get_model().get_sort_column_id()[1]
            cmp_factor = 1
            if this_order == Gtk.SortType.DESCENDING:
                # We need to de-invert the sort order for multi sorting.
                cmp_factor = -1
        i = 0
        while rval == 0 and i < len(self.prev_sort_column_id):
            next_id, next_order = self.prev_sort_column_id[i]
            if next_id == column:
                i += 1
                continue
            next_cmp_factor = cmp_factor * 1
            if next_order == Gtk.SortType.DESCENDING:
                # Set the correct order for multi sorting.
                next_cmp_factor = cmp_factor * -1
            val0 = model.get_value(iter0, next_id)
            val1 = model.get_value(iter1, next_id)
            rval = next_cmp_factor * cmp(val0, val1)
            i += 1
        return rval

    def on_query_tooltip(self, widget, x, y, keyboard_tip, tooltip):
        if not widget.get_tooltip_context(x, y, keyboard_tip):
            return False
        bool, wx, wy, model, path, sel_iter = widget.get_tooltip_context(
            x, y, keyboard_tip
        )
        bin_x, bin_y = widget.convert_widget_to_bin_window_coords(x, y)
        result = widget.get_path_at_pos(bin_x, bin_y)

        if result is not None:
            path, column, cell_x, cell_y = result
            for player, column_number in (
                (self.model[path][0].wplayer, 1),
                (self.model[path][0].bplayer, 3),
            ):
                if column is self.tv.get_column(column_number):
                    tooltip.set_text(get_player_tooltip_text(player, show_status=False))
                    widget.set_tooltip_cell(tooltip, path, None, None)
                    return True
        return False

    def onSelectionChanged(self, selection):
        model, paths = selection.get_selected_rows()
        a_selected_game_is_observable = False
        for path in paths:
            rowiter = model.get_iter(path)
            game = model.get_value(rowiter, 0)
            if not game.private and game.supported:
                a_selected_game_is_observable = True
                break
        self.widgets["observeButton"].set_sensitive(a_selected_game_is_observable)

    def _update_gamesrunning_label(self):
        count = len(self.games)
        self.widgets["gamesRunningLabel"].set_text(_("Games running: %d") % count)

    def onGameAdd(self, games, new_games):
        game_store = {}
        for game in new_games:
            game_store[game] = (
                game,
                self.clearpix,
                game.wplayer.name + game.wplayer.display_titles(),
                game.wplayer.getRatingForCurrentGame(),
                game.bplayer.name + game.bplayer.display_titles(),
                game.bplayer.getRatingForCurrentGame(),
                game.display_text,
                game.display_rated,
            )

        def do_onGameAdd(games, new_games, game_store):
            for game in new_games:
                # game removed before we finish processing "games /bslwBzSLx"
                if game not in game_store:
                    continue
                # log.debug("%s" % game,
                # extra={"task": (self.connection.username, "GTS.onGameAdd")})
                ti = self.store.append(game_store[game])
                self.games[game] = {"ti": ti}
                self.games[game]["private_cid"] = game.connect(
                    "notify::private", self.private_changed
                )
                self._update_gamesrunning_label()

        GLib.idle_add(
            do_onGameAdd, games, new_games, game_store, priority=GLib.PRIORITY_LOW
        )

    def private_changed(self, game, prop):
        try:
            self.store.set(self.games[game]["ti"], 6, game.display_text)
        except KeyError:
            pass
        self.onSelectionChanged(self.tv.get_selection())
        return False

    def onGameRemove(self, games, game):
        def do_onGameRemove(games, game):
            log.debug(
                "%s" % game,
                extra={"task": (self.connection.username, "GTS.onGameRemove")},
            )
            if game not in self.games:
                return
            if self.store.iter_is_valid(self.games[game]["ti"]):
                self.store.remove(self.games[game]["ti"])
            if game.handler_is_connected(self.games[game]["private_cid"]):
                game.disconnect(self.games[game]["private_cid"])
            del self.games[game]
            self._update_gamesrunning_label()

        GLib.idle_add(do_onGameRemove, games, game, priority=GLib.PRIORITY_LOW)

    def onGameObserved(self, bm, game):
        if game in self.games:
            treeiter = self.games[game]["ti"]
            self.store.set_value(treeiter, 1, self.recpix)

    def onGameUnobserved(self, bm, game):
        if game in self.games:
            treeiter = self.games[game]["ti"]
            self.store.set_value(treeiter, 1, self.clearpix)

    def getSelectedPlayer(self):
        model = self.tv.get_model()
        path, col = self.tv.get_cursor()
        col_index = self.tv.get_columns().index(col)
        game = model.get_value(model.get_iter(path), 0)
        return game.bplayer if col_index >= 3 else game.wplayer



from io import StringIO
from urllib.error import URLError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from gi.repository import Gdk, Gtk

from pychess.Utils.const import NAME, UNDOABLE_STATES
from pychess.Savers import pgn
from pychess.widgets import mainwindow

URL = "https://www.chesspastebin.com/api/add/"
APIKEY = "a137d919b75c8766b082367610189358cfb1ba70"


def paste(gamemodel):
    dialog = Gtk.MessageDialog(
        mainwindow(), type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO
    )
    if gamemodel.status in UNDOABLE_STATES:
        text = _(
            "The current game is over. First, please verify the properties of the game."
        )
    else:
        text = _(
            "The current game is not terminated. Its export may have a limited interest."
        )
    text += (
        "\n\n"
        + _("Should %s publicly publish your game as PGN on chesspastebin.com ?") % NAME
    )
    dialog.set_markup(text)
    response = dialog.run()
    dialog.destroy()
    if response != Gtk.ResponseType.YES:
        return

    output = StringIO()
    text = pgn.save(output, gamemodel)
    values = {"apikey": APIKEY, "pgn": text, "name": "PyChess", "sandbox": "false"}

    data = urlencode(values).encode("utf-8")
    req = Request(URL, data)
    try:
        response = urlopen(req, timeout=10)
    except URLError as err:
        if hasattr(err, "reason"):
            print("We failed to reach the server.")
            print("Reason: ", err.reason)
        elif hasattr(err, "code"):
            print("The server couldn't fulfill the request.")
            print("Error code: ", err.code)
    else:
        ID = response.read()
        link = "http://www.chesspastebin.com/?p=%s" % int(ID)
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(link, -1)
        # print(text)
        # print(clipboard.wait_for_text())
        msg_dialog = Gtk.MessageDialog(
            mainwindow(), type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK
        )
        msg = _("Game shared at ") + '<a href="%s">chesspastebin.com</a>' % link
        msg_dialog.set_markup(msg)
        msg_dialog.format_secondary_text(_("(Link is available on clipboard.)"))
        msg_dialog.connect("response", lambda msg_dialog, a: msg_dialog.hide())
        msg_dialog.show()



import asyncio
import os
from os.path import join, dirname, abspath
import sys
import shutil
import json
from functools import partial
from hashlib import md5
from copy import copy
from collections import OrderedDict

from gi.repository import GLib, GObject

from pychess.System import conf
from pychess.System.Log import log
from pychess.System.command import Command
from pychess.System.SubProcess import SubProcess
from pychess.System.prefix import (
    addUserConfigPrefix,
    getDataPrefix,
    getEngineDataPrefix,
)
from pychess.Players.Player import PlayerIsDead
from pychess.Utils import createStoryTextAppEvent
from pychess.Utils.const import (
    BLACK,
    UNKNOWN_REASON,
    WHITE,
    HINT,
    ANALYZING,
    INVERSE_ANALYZING,
    NORMALCHESS,
)
from pychess.Players.CECPEngine import CECPEngine
from pychess.Players.UCIEngine import UCIEngine
from pychess.Players.engineList import PYTHONBIN, VM_LIST, ENGINES_LIST
from pychess.Variants import variants

attrToProtocol = {"uci": UCIEngine, "xboard": CECPEngine}

ENGINE_DEFAULT_LEVEL = 20


class SubProcessError(Exception):
    pass


def md5_sum(filename):
    with open(filename, mode="rb") as file_handle:
        md5sum = md5()
        for buf in iter(partial(file_handle.read, 4096), b""):
            md5sum.update(buf)
    return md5sum.hexdigest()


class EngineDiscoverer(GObject.GObject):
    __gsignals__ = {
        "discovering_started": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "engine_discovered": (GObject.SignalFlags.RUN_FIRST, None, (str, object)),
        "engine_failed": (GObject.SignalFlags.RUN_FIRST, None, (str, object)),
        "all_engines_discovered": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.jsonpath = addUserConfigPrefix("engines.json")
        self.engines = []
        try:
            with open(self.jsonpath) as fh:
                self.engines = json.load(fh)
        except ValueError as err:
            log.warning(
                "engineNest: Couldn't read engines.json, renamed it to .bak\n%s\n%s"
                % (self.jsonpath, err)
            )
            os.rename(self.jsonpath, self.jsonpath + ".bak")
        except OSError as err:
            log.info(
                "engineNest: Couldn't open engines.json, creating a new.\n%s" % err
            )

        # Try to detect engines shipping .eng files on Linux (suggested by HGM on talkchess.com forum)
        for protocol in ("xboard", "uci"):
            for path in ("/usr/local/share/games/plugins", "/usr/share/games/plugins"):
                path = os.path.join(path, protocol)
                if os.path.isdir(path):
                    for entry in os.listdir(path):
                        ext = os.path.splitext(entry)[1]
                        if ext == ".eng":
                            with open(os.path.join(path, entry)) as file_handle:
                                plugin_spec = file_handle.readline().strip()
                                if not plugin_spec.startswith("plugin spec"):
                                    continue

                                engine_command = file_handle.readline().strip()
                                if self.getEngineByName(engine_command) is not None:
                                    continue

                                supported_variants = file_handle.readline().strip()
                                if not supported_variants.startswith("chess"):
                                    continue

                                new_engine = {}
                                if (
                                    engine_command.startswith("cd ")
                                    and engine_command.find(";") > 0
                                ):
                                    parts = engine_command.split(";")
                                    working_directory = parts[0][3:]
                                    engine_command = parts[1]
                                    new_engine["workingDirectory"] = working_directory

                                new_engine["protocol"] = protocol
                                new_engine["name"] = engine_command
                                self.engines.append(new_engine)

        # Initial backup of the loaded engines
        self.backup()

    def __findRunData(self, engine):
        """Searches for a readable, executable named 'name' in the PATH.
        For the PyChess engine, special handling is taken, and we search
        PYTHONPATH as well as the directory from where the 'os' module is
        imported"""
        if engine.get("vm_name") is not None:
            vm_command = engine.get("vm_command")
            altpath = dirname(vm_command) if vm_command else None
            if getattr(sys, "frozen", False) and engine["vm_name"] == "wine":
                vmpath = None
            else:
                vmpath = shutil.which(
                    engine["vm_name"], mode=os.R_OK | os.X_OK, path=altpath
                )

            if engine["name"] == "PyChess.py" and not getattr(sys, "frozen", False):
                path = join(abspath(dirname(__file__)), "PyChess.py")
                if not vmpath.endswith(PYTHONBIN):
                    # from python to python3
                    engine["vm_name"] = PYTHONBIN
                    vmpath = shutil.which(PYTHONBIN, mode=os.R_OK | os.X_OK, path=None)
                if not os.access(path, os.R_OK):
                    path = None
            else:
                command = engine.get("command")
                altpath = dirname(command) if command else None
                path = shutil.which(engine["name"], mode=os.R_OK, path=altpath)

            if vmpath and path:
                return vmpath, path
            elif path and sys.platform == "win32" and engine.get("vm_name") == "wine":
                return None, path

        else:
            command = engine.get("command")
            altpath = dirname(command) if command else None
            if sys.platform == "win32" and not altpath:
                altpath = (
                    os.path.join(getDataPrefix(), "engines")
                    + ";"
                    + os.path.dirname(sys.executable)
                )
            path = shutil.which(
                command if command else engine["name"],
                mode=os.R_OK | os.X_OK,
                path=altpath,
            )
            if path:
                return None, path
        return False

    def __fromUCIProcess(self, subprocess):
        ids = subprocess.ids
        options = subprocess.options
        engine = {}
        if "author" in ids:
            engine["author"] = ids["author"]
        if options:
            engine["options"] = list(options.values())
        return engine

    def __fromCECPProcess(self, subprocess):
        features = subprocess.features
        options = subprocess.options
        engine = {}
        if features["variants"] is not None:
            engine["variants"] = features["variants"].split(",")
        if features["analyze"] == 1:
            engine["analyze"] = True
        if options:
            engine["options"] = list(options.values())
        return engine

    async def __discoverE(self, engine):
        subproc = await self.initEngine(engine, BLACK, False)
        subproc.connect("readyForOptions", self.__discoverE2, engine)
        subproc.prestart()  # Sends the 'start line'

        event = asyncio.Event()
        is_dead = set()
        subproc.start(event, is_dead)

        await event.wait()

        if is_dead:
            # Check if the player died after engine_discovered by our own hands
            if not self.toBeRechecked[engine["name"]][1]:
                log.warning("Engine %s failed discovery" % engine["name"])
                self.emit("engine_failed", engine["name"], engine)
            subproc.kill(UNKNOWN_REASON)

    def __discoverE2(self, subproc, engine):
        if engine.get("protocol") == "uci":
            fresh = self.__fromUCIProcess(subproc)
        elif engine.get("protocol") == "xboard":
            fresh = self.__fromCECPProcess(subproc)
        engine.update(fresh)
        exitcode = subproc.kill(UNKNOWN_REASON)
        if exitcode:
            log.debug("Engine failed %s" % engine["name"])
            self.emit("engine_failed", engine["name"], engine)
            return

        engine["recheck"] = False
        log.debug("Engine finished %s" % engine["name"])
        self.emit("engine_discovered", engine["name"], engine)

    ############################################################################
    # Main loop                                                                #
    ############################################################################

    def __needClean(self, rundata, engine):
        """Check if the filename or md5sum of the engine has changed.
        In that case we need to clean the engine"""
        path = rundata[1]

        # Check if filename is not set, or if it has changed
        if engine.get("command") is None or engine.get("command") != path:
            return True

        # If the engine failed last time, we'll recheck it as well
        if engine.get("recheck"):
            return True

        # Check if md5sum is not set, or if it has changed
        if engine.get("md5") is None:
            return True

        md5sum = md5_sum(path)
        if engine.get("md5") != md5sum:
            return True

        return False

    def __clean(self, rundata, engine):
        """Grab the engine from the referenced engines and attach the attributes
        from rundata. The update engine is ready for discovering.
        """
        vmpath, path = rundata
        md5sum = md5_sum(path)

        # Find the referenced engine
        refeng = self.getReferencedEngine(engine["name"])
        if refeng is None:
            log.warning(
                "Engine '%s' is not in PyChess predefined known engines list"
                % engine.get("name")
            )
            engine["recheck"] = True
        else:
            engine["country"] = refeng["country"]

        # Clean it
        engine["command"] = path
        engine["md5"] = md5sum
        if vmpath is not None:
            engine["vm_command"] = vmpath
        if "variants" in engine:
            del engine["variants"]
        if "options" in engine:
            del engine["options"]

    ######
    # Save the xml
    ######

    def backup(self):
        self.enginesBackup = json.dumps(self.engines)

    def restore(self):
        self.engines = json.loads(self.enginesBackup)

    def hasChanged(self):
        return self.enginesBackup != json.dumps(self.engines)

    def save(self, *args):
        try:
            with open(self.jsonpath, "w") as file_handle:
                json.dump(self.engines, file_handle, indent=1, sort_keys=True)
        except OSError as err:
            log.error(
                "Saving engines.json raised exception: %s"
                % ", ".join(str(a) for a in err.args)
            )

    def pre_discover(self):
        # Remove the expired engines
        self.engines = [
            engine
            for engine in self.engines
            if os.path.isfile(engine.get("command", ""))
        ]

        # Scan the referenced engines to see if they are installed
        for engine in ENGINES_LIST:
            if engine.autoDetect:
                if (
                    self.getEngineByName(engine.name, exactName=False) is not None
                ):  # No rediscovery if already exists
                    continue
                engine = self.getReferencedEngine(engine.name)
                if engine is not None and self.__findRunData(engine):
                    self.engines.append(engine)

        # Check if the existing engines need a refresh
        for engine in self.engines:
            rundata = self.__findRunData(engine)
            if rundata and self.__needClean(rundata, engine):
                self.__clean(rundata, engine)
                engine["recheck"] = True

        # Runs all the engines in toBeRechecked, in order to gather information
        self.toBeRechecked = sorted(
            [(c["name"], [c, False]) for c in self.engines if c.get("recheck")]
        )
        self.toBeRechecked = OrderedDict(self.toBeRechecked)

    def discover(self):
        self.pre_discover()

        def count(self_, name, engine, wentwell):
            if not wentwell:
                print("count() not went well on engine", name)
                self.engines.remove(engine)

            self.toBeRechecked[name][1] = True

            if all([elem[1] for elem in self.toBeRechecked.values()]):
                self.engines.sort(key=lambda x: x["name"])
                self.emit("all_engines_discovered")
                createStoryTextAppEvent("all_engines_discovered")

        self.connect("engine_discovered", count, True)
        self.connect("engine_failed", count, False)
        if self.toBeRechecked:
            self.emit("discovering_started", self.toBeRechecked.keys())
            self.connect("all_engines_discovered", self.save)
            for engine, done in self.toBeRechecked.values():
                if not done:
                    asyncio.create_task(self.__discoverE(engine))
        else:
            self.emit("all_engines_discovered")
            createStoryTextAppEvent("all_engines_discovered")

    ############################################################################
    # Interaction                                                              #
    ############################################################################

    def is_analyzer(self, engine):
        protocol = engine.get("protocol")
        if protocol == "uci":
            return True
        elif protocol == "xboard":
            return engine.get("analyze") is not None

    def getAnalyzers(self):
        return [engine for engine in self.getEngines() if self.is_analyzer(engine)]

    def getEngines(self):
        """Returns a sorted list of engine dicts"""
        return sorted(self.engines, key=lambda engine: engine["name"].lower())

    def getEngineN(self, index):
        return self.getEngines()[index]

    def getEngineByName(self, name, exactName=True):
        approximate = None
        for engine in self.engines:
            if engine["name"] == name:  # Priority to exact name
                return engine
            if not exactName and name.lower() in engine["name"].lower():
                approximate = engine
        return approximate

    def getEngineByMd5(self, md5sum, list=[]):
        if not list:
            list = self.getEngines()
        for engine in list:
            md5_check = engine.get("md5")
            if md5_check is None:
                continue
            if md5_check == md5sum:
                return engine

    def getEngineVariants(self, engine):
        UCI_without_standard_variant = False
        engine_variants = []
        for variantClass in variants.values():
            if variantClass.standard_rules:
                engine_variants.append(variantClass.variant)
            else:
                if engine.get("variants"):
                    if variantClass.cecp_name in engine.get("variants"):
                        engine_variants.append(variantClass.variant)
                # UCI knows Chess960 only
                if engine.get("options"):
                    for option in engine["options"]:
                        if (
                            option["name"] == "UCI_Chess960"
                            and variantClass.cecp_name == "fischerandom"
                        ):
                            engine_variants.append(variantClass.variant)
                        elif option["name"] == "UCI_Variant":
                            UCI_without_standard_variant = (
                                "chess" not in option["choices"]
                            )
                            if (
                                variantClass.cecp_name in option["choices"]
                                or variantClass.cecp_name.lower().replace("-", "")
                                in option["choices"]
                            ):
                                engine_variants.append(variantClass.variant)

        if UCI_without_standard_variant:
            engine_variants.remove(NORMALCHESS)

        return engine_variants

    def getEngineLearn(self):
        # Local helpers
        def has_classical(engine):
            status = engine["protocol"] == "uci"
            if "variants" in engine:
                status = status and "normal" in engine["variants"]
            if "options" in engine:
                for option in engine["options"]:
                    if option["name"] == "UCI_Variant" and "choices" in option:
                        status = status and "chess" in option["choices"]
            return status

        def is_stockfish(engine):
            return "stockfish" in engine["name"].lower()

        # Initialization
        id = conf.get("ana_combobox")
        analyzer_enabled = conf.get("analyzer_check")

        # Stockfish from the preferences
        if (
            analyzer_enabled
        ):  # The analyzer should be active else we might believe that it is irrelevant
            for engine in self.engines:
                if engine["md5"] == id and is_stockfish(engine):
                    return engine["name"]

        # Stockfish from the raw list of engines
        for engine in self.engines:
            if is_stockfish(engine):
                return engine["name"]

        # Preferred engine
        if analyzer_enabled:
            for engine in self.engines:
                if engine["md5"] == id and has_classical(engine):
                    return engine["name"]

        # First found
        for engine in self.engines:
            if has_classical(engine):
                return engine["name"]

        # No engine found
        return None

    def getName(self, engine=None):
        return engine["name"]

    def getReferencedEngine(self, name):
        """This methods builds an automatic entry based on the name of the engine
        and the preconfigured values found for it.
        """
        for engine in ENGINES_LIST:
            if engine.name.lower() in name.lower():
                # Properties of the engine
                result = {
                    "name": os.path.basename(name),
                    "protocol": engine.protocol[:6],
                    "recheck": True,
                    "country": engine.country,
                    "comment": "",
                    "level": (
                        ENGINE_DEFAULT_LEVEL
                        if engine.defaultLevel is None
                        else engine.defaultLevel
                    ),
                }
                if "/" in name or "\\" in name:
                    result["command"] = name
                if engine.protocol == "xboard1":
                    result["protover"] = 1
                if engine.protocol == "xboard2":
                    result["protover"] = 2
                if engine.protocol == "uci":
                    result["analyze"] = True
                # if engine.elo > 0:
                #     result["elo"] = engine.elo

                # Attached interpreter
                ext = os.path.splitext(name.lower())[1]
                if ext != "":
                    for vm in VM_LIST:
                        if ext == vm.ext:
                            result["vm_name"] = vm.name
                            if vm.args is not None:
                                result["vm_args"] = vm.args

                # Linux-specific parameters
                if sys.platform != "win32":
                    if "vm_name" not in result and ext == ".exe":
                        result["vm_name"] = "wine"

                # Verify the host application
                if (
                    result.get("vm_name")
                    and shutil.which(result["vm_name"], mode=os.R_OK | os.X_OK) is None
                ):
                    return None

                # Found result
                return result

        # No result found
        return None

    def getCountry(self, engine):
        return engine.get("country")

    async def initEngine(self, engine, color, lowPriority):
        name = engine["name"]
        protocol = engine["protocol"]
        protover = 2 if engine.get("protover") is None else engine.get("protover")
        path = engine["command"]
        args = [] if engine.get("args") is None else [a for a in engine["args"]]
        if engine.get("vm_command") is not None:
            vmpath = engine["vm_command"]
            vmargs = (
                [] if engine.get("vm_args") is None else [a for a in engine["vm_args"]]
            )
            args = vmargs + [path] + args
            path = vmpath
        md5_engine = engine["md5"]

        working_directory = engine.get("workingDirectory")
        if working_directory:
            workdir = working_directory
        else:
            workdir = getEngineDataPrefix()
        warnwords = ("illegal", "error", "exception")
        try:
            subprocess = SubProcess(
                path,
                args=args,
                warnwords=warnwords,
                cwd=workdir,
                lowPriority=lowPriority,
            )
            await subprocess.start()
        except OSError:
            raise PlayerIsDead
        except asyncio.TimeoutError:
            raise PlayerIsDead
        except GLib.GError:
            raise PlayerIsDead
        except Exception:
            raise PlayerIsDead

        engine_proc = attrToProtocol[protocol](subprocess, color, protover, md5_engine)
        engine_proc.setName(name)

        # If the user has configured special options for this engine, here is
        # where they should be set.

        def optionsCallback(set_option):
            if engine.get("options"):
                for option in engine["options"]:
                    key = option["name"]
                    value = option.get("value")
                    if (value is not None) and option["default"] != value:
                        if protocol == "xboard" and option["type"] == "check":
                            value = int(bool(value))
                        set_option.setOption(key, value)

        engine_proc.connect("readyForOptions", optionsCallback)

        return engine_proc

    async def initPlayerEngine(
        self,
        engine,
        color,
        diffi,
        variant,
        secs=0,
        incr=0,
        moves=0,
        forcePonderOff=False,
    ):
        engine = await self.initEngine(engine, color, False)

        def optionsCallback(engine):
            engine.setOptionStrength(diffi, forcePonderOff)
            engine.setOptionVariant(variant)
            if secs > 0 or incr > 0:
                engine.setOptionTime(secs, incr, moves)

        engine.connect("readyForOptions", optionsCallback)
        engine.prestart()
        return engine

    async def initAnalyzerEngine(self, engine, mode, variant):
        engine = await self.initEngine(engine, WHITE, True)

        def optionsCallback(engine):
            engine.setOptionAnalyzing(mode)
            engine.setOptionVariant(variant)

        engine.connect("readyForOptions", optionsCallback)
        engine.prestart()
        return engine

    def addEngine(self, name, new_engine, protocol, vm_name, vm_args):
        # Default values
        refeng = self.getReferencedEngine(name)
        if refeng is not None:
            engine = copy(refeng)
        else:
            engine = {
                "country": "unknown",
                "comment": "",
                "level": ENGINE_DEFAULT_LEVEL,
            }

        # New values
        engine["name"] = name
        engine["protocol"] = protocol
        engine["command"] = new_engine
        engine["workingDirectory"] = os.path.dirname(new_engine)
        engine["recheck"] = True
        if vm_name is not None:
            engine["vm_name"] = vm_name
        if vm_args is not None:
            engine["vm_args"] = vm_args
        self.engines.append(engine)

    def addEngineFromReference(self, engine):
        if engine is not None:
            self.engines.append(engine)

    def removeEngine(self, name):
        names = [engine["name"] for engine in self.engines]
        index = names.index(name)
        del self.engines[index]


discoverer = EngineDiscoverer()


async def init_engine(analyzer_type, gamemodel, force_engine=None):
    """
    Initializes and starts the engine analyzer of analyzer_type the user has
    configured in the Engines tab of the preferencesDialog, for gamemodel. If no
    such engine is set in the preferences, or if the configured engine doesn't
    support the chess variant being played in gamemodel, then no analyzer is
    started and None is returned.
    """
    if analyzer_type == HINT:
        combo_name = "ana_combobox"
        check_name = "analyzer_check"
        mode = ANALYZING
    else:
        combo_name = "inv_ana_combobox"
        check_name = "inv_analyzer_check"
        mode = INVERSE_ANALYZING

    analyzer = None

    if conf.get(check_name) or force_engine is not None:
        anaengines = list(discoverer.getAnalyzers())
        if len(anaengines) == 0:
            return None

        if force_engine is not None:
            engine = discoverer.getEngineByName(force_engine)
        else:
            engine = discoverer.getEngineByMd5(conf.get(combo_name))

        if engine is None:
            engine = discoverer.getEngineByName(discoverer.getEngineLearn())

        if engine is None:
            engine = anaengines[-1]

        if gamemodel.variant.variant in discoverer.getEngineVariants(engine):
            analyzer = await discoverer.initAnalyzerEngine(
                engine, mode, gamemodel.variant
            )
            log.debug(f"{analyzer_type} analyzer: {repr(analyzer)}")

    return analyzer


def is_uci(engine_command):
    command = Command(engine_command, "uci\nquit\n")
    status, output, err = command.run(timeout=5)
    uci = False
    for line in output.splitlines():
        line = line.rstrip()
        if line == "uciok" or line.startswith("info string"):
            uci = True
            break
        elif "Error" in line or "Illegal" in line or "Invalid" in line:
            break
    return uci


def is_cecp(engine_command):
    command = Command(engine_command, "xboard\nprotover 2\nquit\n")
    status, output, err = command.run(timeout=5)
    cecp = False
    for line in output.splitlines():
        line = line.rstrip()
        if "feature" in line and "done" in line:
            cecp = True
            break
        elif "Error" in line or "Illegal" in line or "Invalid" in line:
            break
    return cecp


if __name__ == "__main__":
    loop = asyncio.get_event_loop()

    def discovering_started(discoverer, names):
        print("discovering_started", names)

    discoverer.connect("discovering_started", discovering_started)

    def engine_discovered(discoverer, name, engine):
        sys.stdout.write(".")

    discoverer.connect("engine_discovered", engine_discovered)

    def all_engines_discovered(discoverer):
        print("all_engines_discovered")
        print([engine["name"] for engine in discoverer.getEngines()])
        loop.stop()

    discoverer.connect("all_engines_discovered", all_engines_discovered)

    discoverer.discover()

    try:
        loop.run_forever()
    except Exception:
        print(sys.exc_info())
        sys.exit(0)



from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.leval import clearPawnTable
from pychess.Utils.lutils.lmove import listToSan
from pychess.Utils.lutils import lsearch
from pychess.Utils.const import NORMALCHESS
import sys
from time import time

# For now, we use the benchmark positions from Stockfish.
benchmarkPositions = [
    "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1",
    "r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq - 0 10",
    "8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - - 0 11",
    "4rrk1/pp1n3p/3q2pQ/2p1pb2/2PP4/2P3N1/P2B2PP/4RRK1 b - - 7 19",
    "rq3rk1/ppp2ppp/1bnpb3/3N2B1/3NP3/7P/PPPQ1PP1/2KR3R w - - 7 14",
    "r1bq1r1k/1pp1n1pp/1p1p4/4p2Q/4Pp2/1BNP4/PPP2PPP/3R1RK1 w - - 2 14",
    "r3r1k1/2p2ppp/p1p1bn2/8/1q2P3/2NPQN2/PPP3PP/R4RK1 b - - 2 15",
    "r1bbk1nr/pp3p1p/2n5/1N4p1/2Np1B2/8/PPP2PPP/2KR1B1R w kq - 0 13",
    "r1bq1rk1/ppp1nppp/4n3/3p3Q/3P4/1BP1B3/PP1N2PP/R4RK1 w - - 1 16",
    "4r1k1/r1q2ppp/ppp2n2/4P3/5Rb1/1N1BQ3/PPP3PP/R5K1 w - - 1 17",
    "2rqkb1r/ppp2p2/2npb1p1/1N1Nn2p/2P1PP2/8/PP2B1PP/R1BQK2R b KQ - 0 11",
    "r1bq1r1k/b1p1npp1/p2p3p/1p6/3PP3/1B2NN2/PP3PPP/R2Q1RK1 w - - 1 16",
    "3r1rk1/p5pp/bpp1pp2/8/q1PP1P2/b3P3/P2NQRPP/1R2B1K1 b - - 6 22",
    "r1q2rk1/2p1bppp/2Pp4/p6b/Q1PNp3/4B3/PP1R1PPP/2K4R w - - 2 18",
    "4k2r/1pb2ppp/1p2p3/1R1p4/3P4/2r1PN2/P4PPP/1R4K1 b - - 3 22",
    "3q2k1/pb3p1p/4pbp1/2r5/PpN2N2/1P2P2P/5PP1/Q2R2K1 b - - 4 26",
]


def benchmark(maxdepth=6):
    """Times a search of a static list of positions."""

    suite_time = time()
    suite_nodes = lsearch.nodes
    lsearch.endtime = sys.maxsize
    lsearch.searching = True
    for i, fen in enumerate(benchmarkPositions):
        lsearch.table.clear()
        clearPawnTable()
        board = LBoard(NORMALCHESS)
        board.applyFen(fen)
        pos_start_time = time()
        pos_start_nodes = lsearch.nodes
        for depth in range(1, maxdepth):
            mvs, scr = lsearch.alphaBeta(board, depth)
            pos_time = time() - pos_start_time
            pos_nodes = lsearch.nodes - pos_start_nodes
            pv = " ".join(listToSan(board, mvs))
            time_cs = int(100 * pos_time)
            print(depth, scr, time_cs, pos_nodes, pv)
        print(
            "Searched position",
            i,
            "at",
            int(pos_nodes / pos_time) if pos_time > 0 else pos_nodes,
            "n/s",
        )
    suite_time = time() - suite_time
    suite_nodes = lsearch.nodes - suite_nodes
    print(
        "Total:",
        suite_nodes,
        "nodes in",
        suite_time,
        "s: ",
        suite_nodes / suite_time,
        "n/s",
    )
    lsearch.nodes = 0



# The purpose of this module, is to give a certain position a score.
# The greater the score, the better the position

from pychess.Utils.const import (
    WHITE,
    BLACK,
    LOSERSCHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    ASEAN_VARIANTS,
    ATOMICCHESS,
    CRAZYHOUSECHESS,
    RACINGKINGSCHESS,
    THREECHECKCHESS,
    BPAWN,
    BISHOP,
    KNIGHT,
    QUEEN,
    KING,
    PAWN,
    ROOK,
    CAS_FLAGS,
    H7,
    B6,
    A7,
    H2,
    G3,
    A2,
    B3,
    G6,
    D1,
    G8,
    B8,
    G1,
    B1,
)
from .bitboard import iterBits, firstBit, lsb
from .ldata import (
    fileBits,
    bitPosArray,
    PIECE_VALUES,
    FILE,
    RANK,
    PAWN_VALUE,
    WHITE_SQUARES,
    BLACK_SQUARES,
    ASEAN_PIECE_VALUES,
    ATOMIC_PIECE_VALUES,
    CRAZY_PIECE_VALUES,
    kwingpawns1,
    kwingpawns2,
    qwingpawns1,
    qwingpawns2,
    frontWall,
    endingKing,
    brank7,
    brank8,
    distance,
    isolaniMask,
    d2e2,
    passedScores,
    squarePawnMask,
    moveArray,
    brank67,
    lbox,
    stonewall,
    isolani_normal,
    isolani_weaker,
    passedPawnMask,
    fromToRay,
    pawnScoreBoard,
    sdistance,
    taxicab,
    racingKing,
)
from .lsort import staticExchangeEvaluate
from .lmovegen import newMove
from pychess.Variants.threecheck import checkCount
from pychess.Variants.racingkings import testKingInEightRow
from ctypes import create_string_buffer, memset
from struct import Struct

# from random import randint
randomval = 0  # randint(8,12)/10.

CHECK_BONUS = (0, 100, 500, 2000)


def evaluateComplete(board, color):
    """A detailed evaluation function, taking into account
    several positional factors"""

    s, phase = evalMaterial(board, color)
    if board.variant in (LOSERSCHESS, SUICIDECHESS, GIVEAWAYCHESS):
        return s

    s += evalKing(board, color, phase) - evalKing(board, 1 - color, phase)
    if board.variant == RACINGKINGSCHESS or board.variant == THREECHECKCHESS:
        return s

    s += evalBishops(board, color, phase) - evalBishops(board, 1 - color, phase)
    s += evalRooks(board, color, phase) - evalRooks(board, 1 - color, phase)
    s += evalDoubleQR7(board, color, phase) - evalDoubleQR7(board, 1 - color, phase)
    s += evalKingTropism(board, color, phase) - evalKingTropism(board, 1 - color, phase)
    if board.variant in ASEAN_VARIANTS:
        return s
    s += evalDev(board, color, phase) - evalDev(board, 1 - color, phase)
    if board.variant == ATOMICCHESS:
        return s
    pawnScore, passed, weaked = cacheablePawnInfo(board, phase)
    s += pawnScore if color == WHITE else -pawnScore
    s += evalPawnStructure(board, color, phase, passed, weaked) - evalPawnStructure(
        board, 1 - color, phase, passed, weaked
    )

    s += evalTrappedBishops(board, color)
    s += randomval

    return s


################################################################################
# evalMaterial                                                                 #
################################################################################


def evalMaterial(board, color):
    # While opp reached rank 8 we can save the game if we also reach it
    # but for this we have to force the shortest (one king move) draw line!
    if board.variant == RACINGKINGSCHESS and testKingInEightRow(board):
        return [0, 0]

    pieceCount = board.pieceCount
    opcolor = 1 - color
    material = [0, 0]
    if board.variant == CRAZYHOUSECHESS:
        for piece in range(PAWN, KING):
            material[WHITE] += CRAZY_PIECE_VALUES[piece] * pieceCount[WHITE][piece]
            material[BLACK] += CRAZY_PIECE_VALUES[piece] * pieceCount[BLACK][piece]
            material[WHITE] += CRAZY_PIECE_VALUES[piece] * board.holding[WHITE][piece]
            material[BLACK] += CRAZY_PIECE_VALUES[piece] * board.holding[BLACK][piece]
    elif board.variant == LOSERSCHESS:
        for piece in range(PAWN, KING):
            material[WHITE] += pieceCount[WHITE][piece]
            material[BLACK] += pieceCount[BLACK][piece]
    elif board.variant == SUICIDECHESS or board.variant == GIVEAWAYCHESS:
        for piece in range(PAWN, KING + 1):
            material[WHITE] += pieceCount[WHITE][piece]
            material[BLACK] += pieceCount[BLACK][piece]
    elif board.variant == ATOMICCHESS:
        for piece in range(PAWN, KING + 1):
            material[WHITE] += ATOMIC_PIECE_VALUES[piece] * pieceCount[WHITE][piece]
            material[BLACK] += ATOMIC_PIECE_VALUES[piece] * pieceCount[BLACK][piece]
    elif board.variant in ASEAN_VARIANTS:
        for piece in range(PAWN, KING + 1):
            material[WHITE] += ASEAN_PIECE_VALUES[piece] * pieceCount[WHITE][piece]
            material[BLACK] += ASEAN_PIECE_VALUES[piece] * pieceCount[BLACK][piece]
    else:
        for piece in range(PAWN, KING):
            material[WHITE] += PIECE_VALUES[piece] * pieceCount[WHITE][piece]
            material[BLACK] += PIECE_VALUES[piece] * pieceCount[BLACK][piece]

    phase = max(1, 8 - (material[WHITE] + material[BLACK]) // 1150)

    # If both sides are equal, we don't need to compute anything!
    if material[BLACK] == material[WHITE]:
        return 0, phase

    matTotal = sum(material)

    # Who is leading the game, material-wise?
    if material[color] > material[opcolor]:
        leading = color
    else:
        leading = opcolor

    if board.variant in (LOSERSCHESS, SUICIDECHESS, GIVEAWAYCHESS):
        val = material[leading] - material[1 - leading]
        val = int(100 * PAWN_VALUE * val / max(material[WHITE], material[BLACK]))
        if leading == 1 - color:
            return val, phase
        return -val, phase

    pawns = pieceCount[leading][PAWN]
    matDiff = material[leading] - material[1 - leading]
    val = min(2400, matDiff) + (matDiff * (12000 - matTotal) * pawns) // (
        6400 * (pawns + 1)
    )

    if leading == color:
        return val, phase
    return -val, phase

    ################################################################################
    # evalKingTropism                                                              #
    ################################################################################


pawnTropism = [[0] * 64 for i in range(64)]
bishopTropism = [[0] * 64 for i in range(64)]
knightTropism = [[0] * 64 for i in range(64)]
rookTropism = [[0] * 64 for i in range(64)]
queenTropism = [[0] * 64 for i in range(64)]

for pcord in range(64):
    for kcord in range(pcord + 1, 64):
        pawnTropism[pcord][kcord] = pawnTropism[kcord][pcord] = (
            (14 - taxicab[pcord][kcord]) ** 2 * 10 / 169
        )
        knightTropism[pcord][kcord] = knightTropism[kcord][pcord] = (
            6 - distance[KNIGHT][pcord][kcord]
        ) ** 2 * 2
        bishopTropism[pcord][kcord] = bishopTropism[kcord][pcord] = (
            (14 - distance[BISHOP][pcord][kcord] * sdistance[pcord][kcord]) ** 2
            * 30
            // 169
        )
        rookTropism[pcord][kcord] = rookTropism[kcord][pcord] = (
            (14 - distance[ROOK][pcord][kcord] * sdistance[pcord][kcord]) ** 2
            * 40
            // 169
        )
        queenTropism[pcord][kcord] = queenTropism[kcord][pcord] = (
            (14 - distance[QUEEN][pcord][kcord] * sdistance[pcord][kcord]) ** 2
            * 50
            // 169
        )

tropisms = {
    PAWN: pawnTropism,
    KNIGHT: knightTropism,
    BISHOP: bishopTropism,
    ROOK: rookTropism,
    QUEEN: queenTropism,
}


def evalKingTropism(board, color, phase):
    """All other things being equal, having your Knights, Queens and Rooks
    close to the opponent's king is a good thing"""
    _tropisms = tropisms
    _lsb = lsb
    opcolor = 1 - color
    pieces = board.boards[color]

    opking = board.kings[opcolor]

    score = 0
    for piece in range(KNIGHT, KING):
        #    for piece in range(PAWN, KING):
        bitboard = pieces[piece]
        tropism = _tropisms[piece]
        # inlined iterBits()
        while bitboard:
            bit = bitboard & -bitboard
            score += tropism[_lsb[bit]][opking]
            bitboard -= bit
    return score


################################################################################
# evalPawnStructure                                                            #
################################################################################

# For pawn hash, don't use buckets. Store:
# key         high 16 bits of pawn hash key
# score       score from white's point of view
# passed      bitboard of passed pawns
# weaked      bitboard of weak pawns
pawnEntryType = Struct("=H h Q Q")
PAWN_HASH_SIZE = 16384
PAWN_PHASE_KEY = (0x343D, 0x055D, 0x3D3C, 0x1A1C, 0x28AA, 0x19EE, 0x1538, 0x2A99)
pawntable = create_string_buffer(PAWN_HASH_SIZE * pawnEntryType.size)


def clearPawnTable():
    memset(pawntable, 0, PAWN_HASH_SIZE * pawnEntryType.size)


def probePawns(board, phase):
    index = (board.pawnhash % PAWN_HASH_SIZE) ^ PAWN_PHASE_KEY[phase - 1]
    key, score, passed, weaked = pawnEntryType.unpack_from(
        pawntable, index * pawnEntryType.size
    )
    if key == (board.pawnhash >> 14) & 0xFFFF:
        return score, passed, weaked
    return None


def recordPawns(board, phase, score, passed, weaked):
    index = (board.pawnhash % PAWN_HASH_SIZE) ^ PAWN_PHASE_KEY[phase - 1]
    key = (board.pawnhash >> 14) & 0xFFFF
    pawnEntryType.pack_into(
        pawntable, index * pawnEntryType.size, key, score, passed, weaked
    )


def cacheablePawnInfo(board, phase):
    entry = probePawns(board, phase)
    if entry:
        return entry

    score = 0
    passed = 0
    weaked = 0

    for color in WHITE, BLACK:
        opcolor = 1 - color
        pawns = board.boards[color][PAWN]
        oppawns = board.boards[opcolor][PAWN]

        nfile = [0] * 8
        pScoreBoard = pawnScoreBoard[color]
        for cord in iterBits(pawns):
            score += pScoreBoard[cord] * 2

            # Passed pawns
            if not oppawns & passedPawnMask[color][cord]:
                if (color == WHITE and not fromToRay[cord][cord | 56] & pawns) or (
                    color == BLACK and not fromToRay[cord][cord & 7] & pawns
                ):
                    passed |= bitPosArray[cord]
                    score += (passedScores[color][cord >> 3] * phase) // 12

            # Backward pawns
            backward = False

            if color == WHITE:
                i = cord + 8
            else:
                i = cord - 8

            ptype = color == WHITE and PAWN or BPAWN
            opptype = color == BLACK and PAWN or BPAWN

            if (
                not (passedPawnMask[opcolor][i] & ~fileBits[cord & 7] & pawns)
                and board.arBoard[i] != PAWN
            ):
                n1 = bin(pawns & moveArray[opptype][i]).count("1")
                n2 = bin(oppawns & moveArray[ptype][i]).count("1")
                if n1 < n2:
                    backward = True

            if not backward and bitPosArray[cord] & brank7[opcolor]:
                i = i + (color == WHITE and 8 or -8)
                if not (passedPawnMask[opcolor][i] & ~fileBits[1] & pawns):
                    n1 = bin(pawns & moveArray[opptype][i]).count("1")
                    n2 = bin(oppawns & moveArray[ptype][i]).count("1")
                    if n1 < n2:
                        backward = True

                if not backward and bitPosArray[cord] & brank7[opcolor]:
                    i = i + (color == WHITE and 8 or -8)
                    if not (passedPawnMask[opcolor][i] & ~fileBits[1] & pawns):
                        n1 = bin(pawns & moveArray[opptype][i]).count("1")
                        n2 = bin(oppawns & moveArray[ptype][i]).count("1")
                        if n1 < n2:
                            backward = True

            if backward:
                weaked |= bitPosArray[cord]
                score += -(8 + phase)  # Backward pawn penalty

            # Pawn base under attack
            if moveArray[ptype][cord] & oppawns and moveArray[ptype][cord] & pawns:
                score += -18

            # Increment file count for isolani & doubled pawn evaluation
            nfile[cord & 7] += 1

        for i in range(8):
            # Doubled pawns
            if nfile[i] > 1:
                score += -(8 + phase)

            # Isolated pawns
            if nfile[i] and not pawns & isolaniMask[i]:
                if not fileBits[i] & oppawns:
                    # Isolated on a half-open file
                    score += isolani_weaker[i] * nfile[i]
                else:
                    # Normal isolated pawn
                    score += isolani_normal[i] * nfile[i]
                weaked |= pawns & fileBits[i]

        # Penalize having eight pawns
        if board.pieceCount[color][PAWN] == 8:
            score -= 10

        # Detect stonewall formation in our pawns
        if stonewall[color] & pawns == stonewall[color]:
            score += 10

        # Penalize Locked pawns
        n = bin((pawns >> 8) & oppawns & lbox).count("1")
        score -= n * 10

        # Switch point of view when switching colors
        score = -score

    recordPawns(board, phase, score, passed, weaked)
    return score, passed, weaked


def evalPawnStructure(board, color, phase, passed, weaked):
    """
    Pawn evaluation is based on the following factors:
    1.  Pawn square tables.
    2.  Passed pawns.
    3.  Backward pawns.
    4.  Pawn base under attack.
    5.  Doubled pawns
    6.  Isolated pawns
    7.  Connected passed pawns on 6/7th rank.
    8.  Unmoved & blocked d, e pawn
    9.  Passed pawn which cannot be caught.
    10. Pawn storms.
    Notice: The function has better precicion for current player
    """

    boards = board.boards[color]

    if not boards[PAWN]:
        return 0

    king = board.kings[color]
    pawns = boards[PAWN]

    opcolor = 1 - color
    opking = board.kings[opcolor]
    opboards = board.boards[opcolor]

    score = 0
    passed &= pawns
    weaked &= pawns

    # This section of the pawn code cannot be saved into the pawn hash as
    # they depend on the position of other pieces.  So they have to be
    # calculated again.
    if passed:
        # Connected passed pawns on 6th or 7th rank
        t = passed & brank67[color]
        opMajorCount = 0
        for p in range(KNIGHT, KING):
            opMajorCount += board.pieceCount[opcolor][p]
        if t and opMajorCount == 1:
            n1 = FILE(opking)
            n2 = RANK(opking)
            for f in range(7):
                if (
                    t & fileBits[f]
                    and t & fileBits[f + 1]
                    and (
                        n1 < f - 1
                        or n1 > f + 1
                        or (color == WHITE and n2 < 4)
                        or (color == BLACK and n2 > 3)
                    )
                ):
                    score += 50

            # Enemy has no pieces & King is outcolor of passed pawn square
        if not opMajorCount:
            for cord in iterBits(passed):
                if board.color == color:
                    if not squarePawnMask[color][cord] & opboards[KING]:
                        score += passedScores[color][RANK(cord)]
                else:
                    if not moveArray[KING][opking] & squarePawnMask[color][cord]:
                        score += passedScores[color][RANK(cord)]

        # Estimate if any majors are able to hunt us down
        for pawn in iterBits(passed):
            found_hunter = False
            if color == WHITE:
                prom_cord = 7 << 3 | FILE(pawn)
            else:
                prom_cord = FILE(pawn)
            distance_to_promotion = distance[PAWN][pawn][prom_cord]
            for piece in range(KNIGHT, KING + 1):
                for cord in iterBits(opboards[piece]):
                    hunter_distance = distance[piece][cord][prom_cord]
                    if hunter_distance <= distance_to_promotion:
                        found_hunter = True
                        break
                if found_hunter:
                    break
            if not found_hunter:
                score += passedScores[color][RANK(pawn)] // 5

    # Penalize Pawn on d2,e2/d7,e7 is blocked
    blocker = board.blocker
    if color == WHITE and ((pawns & d2e2[WHITE]) >> 8) & blocker:
        score -= 48
    elif color == BLACK and ((pawns & d2e2[BLACK]) << 8) & blocker:
        score -= 48

    # If both colors are castled on different colors, bonus for pawn storms
    if abs(FILE(king) - FILE(opking)) >= 4 and phase < 6:
        n1 = FILE(opking)
        p = (isolaniMask[n1] | fileBits[n1]) & pawns
        score += sum(10 * (5 - distance[KING][c][opking]) for c in iterBits(p))

    return score


# evalBateries
def evalDoubleQR7(board, color, phase):
    """Tests for QR, RR, QB and BB combos on the 7th rank. These are dangerous
    to kings, and good at killing pawns"""

    opcolor = 1 - board.color
    boards = board.boards[color]
    opboards = board.boards[opcolor]

    if bin((boards[QUEEN] | boards[ROOK]) & brank7[color]).count("1") >= 2 and (
        opboards[KING] & brank8[color] or opboards[PAWN] & brank7[color]
    ):
        return 30

    return 0


def evalKing(board, color, phase):
    # Should avoid situations like those:
    # r - - - n K - -
    # which makes forks more easy
    # and
    # R - - - K - - -
    # and
    # - - - - - - - -
    # - - - K - - - -
    # - - - - - - - -
    # - - - - - - - -
    # - - - - - - B -
    # which might turn bad

    # Also being check should be avoided, like
    # - q - - - K - r
    # and
    # - - - - - n - -
    # - - - K - - - R

    king = board.kings[color]

    if board.variant == RACINGKINGSCHESS:
        return racingKing[king]

    if board.variant == THREECHECKCHESS:
        return CHECK_BONUS[min(3, checkCount(board, 1 - color))]

    # If we are in endgame, we want our king in the center, and theirs far away
    if phase >= 6:
        return endingKing[king]

    # else if castled, prefer having some pawns in front
    elif FILE(king) not in (3, 4) and RANK(king) in (0, 8):
        if color == WHITE:
            if FILE(king) < 3:
                wall1 = frontWall[color][B1]
            else:
                wall1 = frontWall[color][G1]
            wall2 = wall1 >> 8
        else:
            if FILE(king) < 3:
                wall1 = frontWall[color][B8]
            else:
                wall1 = frontWall[color][G8]
            wall2 = wall1 << 8

        pawns = board.boards[color][PAWN]
        total_in_front = bin(wall1 | wall2 & pawns).count("1")
        numbermod = (0, 3, 6, 9, 7, 5, 3)[total_in_front]

        s = bin(wall1 & pawns).count("1") * 2 + bin(wall2 & pawns).count("1")
        return (s * numbermod * 5) // 6

    return 0


def evalDev(board, color, phase):
    """
    Calculate the development score for side (for opening only).
    Penalize the following.
    .  Uncastled and cannot castled
    .  Early queen move.
    -  bad wing pawns
    """

    # If we are castled or beyond the 20th move, no more evalDev

    if board.plyCount >= 38:
        return 0

    score = 0

    if not board.hasCastled[color]:
        boards = board.boards[color]
        pawns = boards[PAWN]

        # We don't encourage castling, but it should always be possible
        if not board.castling & CAS_FLAGS[color][0]:
            score -= 40
        if not board.castling & CAS_FLAGS[color][1]:
            score -= 50

        # Should keep queen home
        cord = firstBit(boards[QUEEN])
        if cord != D1 + 56 * color:
            score -= 30

        qpawns = max(qwingpawns1[color] & pawns, qwingpawns2[color] & pawns)
        kpawns = max(kwingpawns1[color] & pawns, kwingpawns2[color] & pawns)

        if qpawns != 2 and kpawns != 2:
            # Structure destroyed in both sides
            score -= 35
        else:
            # Discourage any wing pawn moves
            score += (qpawns + kpawns) * 6

    return score


def evalBishops(board, color, phase):
    opcolor = 1 - color
    bishops = board.boards[color][BISHOP]
    if not bishops:
        return 0

    pawns = board.boards[color][PAWN]
    oppawns = board.boards[opcolor][PAWN]

    score = 0

    # Avoid having too many pawns on you bishop's color.
    # In late game phase, add a bonus for enemy pieces on your bishop's color.

    if board.pieceCount[color][BISHOP] == 1:
        squareMask = WHITE_SQUARES if (bishops & WHITE_SQUARES) else BLACK_SQUARES
        score = (
            -bin(pawns & squareMask).count("1")
            - bin(oppawns & squareMask).count("1") // 2
        )
        if phase > 6:
            score += bin(board.friends[1 - color] & squareMask).count("1")

    return score


def evalTrappedBishops(board, color):
    """Check for bishops trapped at A2/H2/A7/H7"""

    _bitPosArray = bitPosArray
    wbishops = board.boards[WHITE][BISHOP]
    bbishops = board.boards[BLACK][BISHOP]
    wpawns = board.boards[WHITE][PAWN]
    bpawns = board.boards[BLACK][PAWN]
    score = 0

    if bbishops:
        if bbishops & _bitPosArray[A2] and wpawns & _bitPosArray[B3]:
            see = staticExchangeEvaluate(board, newMove(A2, B3))
            if see < 0:
                score -= see
        if bbishops & _bitPosArray[H2] and wpawns & _bitPosArray[G3]:
            see = staticExchangeEvaluate(board, newMove(H2, G3))
            if see < 0:
                score -= see

    if wbishops:
        if wbishops & _bitPosArray[A7] and bpawns & _bitPosArray[B6]:
            see = staticExchangeEvaluate(board, newMove(A7, B6))
            if see < 0:
                score += see
        if wbishops & _bitPosArray[H7] and bpawns & _bitPosArray[G6]:
            see = staticExchangeEvaluate(board, newMove(H7, G6))
            if see < 0:
                score += see

    return score if color == WHITE else -score


def evalRooks(board, color, phase):
    """rooks on open/half-open files"""

    opcolor = 1 - color
    boards = board.boards[color]
    rooks = boards[ROOK]

    if not rooks:
        return 0

    opking = board.kings[opcolor]
    score = 0

    if phase < 7:
        for cord in iterBits(rooks):
            file = cord & 7
            if not boards[PAWN] & fileBits[file]:
                if file == 5 and opking & 7 >= 4:
                    score += 40
                score += 5
                if not boards[PAWN] & fileBits[file]:
                    score += 6

    return score



from gi.repository import Gtk, Gdk, GObject

from pychess.System import conf
from pychess.Utils.Cord import Cord
from pychess.Utils.Move import Move, parseAny, toAN
from pychess.Utils.const import (
    ARTIFICIAL,
    FLAG_CALL,
    ABORT_OFFER,
    LOCAL,
    TAKEBACK_OFFER,
    ADJOURN_OFFER,
    DRAW_OFFER,
    RESIGNATION,
    HURRY_ACTION,
    PAUSE_OFFER,
    RESUME_OFFER,
    RUNNING,
    DROP,
    DROP_VARIANTS,
    PAWN,
    QUEEN,
    KING,
    SITTUYINCHESS,
    QUEEN_PROMOTION,
    KNIGHT_PROMOTION,
    SCHESS,
    HAWK,
    ELEPHANT,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
    LIGHTBRIGADECHESS,
    WHITE,
)

from pychess.Utils.logic import validate
from pychess.Utils.lutils.bitboard import iterBits
from pychess.Utils.lutils import lmove, lmovegen
from pychess.Utils.lutils.lmove import ParsingError

from . import preferencesDialog
from .GatingDialog import GatingDialog
from .PromotionDialog import PromotionDialog
from .BoardView import BoardView, rect, join


class BoardControl(Gtk.EventBox):
    """Creates a BoardView for GameModel to control move selection,
    action menu selection and emits signals to let Human player
    make moves and emit offers.
    SetuPositionDialog uses setup_position=True to disable most validation.
    When game_preview=True just do circles and arrows
    """

    __gsignals__ = {
        "shapes_changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "piece_moved": (GObject.SignalFlags.RUN_FIRST, None, (object, int)),
        "action": (GObject.SignalFlags.RUN_FIRST, None, (str, object, object)),
    }

    def __init__(
        self, gamemodel, action_menu_items, setup_position=False, game_preview=False
    ):
        GObject.GObject.__init__(self)
        self.setup_position = setup_position
        self.game_preview = game_preview

        self.view = BoardView(gamemodel, setup_position=setup_position)

        self.add(self.view)
        self.variant = gamemodel.variant
        self.gatingDialog = GatingDialog(SCHESS)
        self.promotionDialog = PromotionDialog(self.variant.variant)

        self.RANKS = gamemodel.boards[0].RANKS
        self.FILES = gamemodel.boards[0].FILES

        self.action_menu_items = action_menu_items
        self.connections = {}
        for key, menuitem in self.action_menu_items.items():
            if menuitem is None:
                print(key)
            # print("...connect to", key, menuitem)
            self.connections[menuitem] = menuitem.connect(
                "activate", self.actionActivate, key
            )
        self.view_cid = self.view.connect("shownChanged", self.shownChanged)

        self.gamemodel = gamemodel
        self.gamemodel_cids = []
        self.gamemodel_cids.append(
            gamemodel.connect("moves_undoing", self.moves_undone)
        )
        self.gamemodel_cids.append(gamemodel.connect("game_ended", self.game_ended))
        self.gamemodel_cids.append(gamemodel.connect("game_started", self.game_started))

        self.cids = []
        self.cids.append(self.connect("button_press_event", self.button_press))
        self.cids.append(self.connect("button_release_event", self.button_release))
        self.add_events(
            Gdk.EventMask.LEAVE_NOTIFY_MASK | Gdk.EventMask.POINTER_MOTION_MASK
        )
        self.cids.append(self.connect("motion_notify_event", self.motion_notify))
        self.cids.append(self.connect("leave_notify_event", self.leave_notify))

        self.selected_last = None
        self.normalState = NormalState(self)
        self.selectedState = SelectedState(self)
        self.activeState = ActiveState(self)
        self.lockedNormalState = LockedNormalState(self)
        self.lockedSelectedState = LockedSelectedState(self)
        self.lockedActiveState = LockedActiveState(self)
        self.currentState = self.normalState

        self.lockedPly = self.view.shown
        self.possibleBoards = {self.lockedPly: self._genPossibleBoards(self.lockedPly)}

        self.allowPremove = False

        def onGameStart(gamemodel):
            if not self.setup_position:
                for player in gamemodel.players:
                    if player.__type__ == LOCAL:
                        self.allowPremove = True

        self.gamemodel_cids.append(gamemodel.connect("game_started", onGameStart))
        self.keybuffer = ""

        self.pre_arrow_from = None
        self.pre_arrow_to = None

    def _del(self):
        self.view.disconnect(self.view_cid)
        for cid in self.cids:
            self.disconnect(cid)

        for obj, conid in self.connections.items():
            # print("...disconnect from ", obj)
            obj.disconnect(conid)
        self.connections = {}
        self.action_menu_items = {}

        for cid in self.gamemodel_cids:
            self.gamemodel.disconnect(cid)

        self.view._del()

        self.promotionDialog = None

        self.normalState = None
        self.selectedState = None
        self.activeState = None
        self.lockedNormalState = None
        self.lockedSelectedState = None
        self.lockedActiveState = None
        self.currentState = None

    def getGating(self, castling, hawk, elephant):
        color = self.view.model.boards[-1].color
        gating = self.gatingDialog.runAndHide(color, castling, hawk, elephant)
        return gating

    def getPromotion(self):
        color = self.view.model.boards[-1].color
        variant = self.view.model.boards[-1].variant
        promotion = self.promotionDialog.runAndHide(color, variant)
        return promotion

    def play_sound(self, move, board):
        if move.is_capture(board):
            sound = "aPlayerCaptures"
        else:
            sound = "aPlayerMoves"

        if board.board.isChecked():
            sound = "aPlayerChecks"

        preferencesDialog.SoundTab.playAction(sound)

    def play_or_add_move(self, board, move):
        if board.board.next is None:
            # at the end of variation or main line
            if not self.view.shownIsMainLine():
                # add move to existing variation
                self.view.model.add_move2variation(
                    board, move, self.view.shown_variation_idx
                )
                self.view.showNext()
            else:
                # create new variation
                new_vari = self.view.model.add_variation(board, [move])
                self.view.setShownBoard(new_vari[-1])
        else:
            # inside variation or main line
            if board.board.next.lastMove == move.move:
                # replay mainline move
                if self.view.model.lesson_game:
                    next_board = self.view.model.getBoardAtPly(
                        self.view.shown + 1, self.view.shown_variation_idx
                    )
                    self.play_sound(move, board)
                    incr = (
                        1
                        if len(
                            self.view.model.variations[self.view.shown_variation_idx]
                        )
                        - 1
                        == board.ply - self.view.model.lowply + 1
                        else 2
                    )
                    if incr == 2:
                        next_next_board = self.view.model.getBoardAtPly(
                            self.view.shown + 2, self.view.shown_variation_idx
                        )
                        # If there is any opponent move variation let the user choose opp next move
                        if any(
                            child
                            for child in next_next_board.board.children
                            if isinstance(child, list)
                        ):
                            self.view.infobar.opp_turn()
                            self.view.showNext()
                        # If there is some comment to read let the user read it before opp move
                        elif any(
                            child
                            for child in next_board.board.children
                            if isinstance(child, str)
                        ):
                            self.view.infobar.opp_turn()
                            self.view.showNext()

                        # If there is nothing to wait for we make opp next move
                        else:
                            self.view.showNext()
                            self.view.infobar.your_turn()
                            self.view.showNext()
                    else:
                        if self.view.shownIsMainLine():
                            preferencesDialog.SoundTab.playAction("puzzleSuccess")
                            self.view.infobar.get_next_puzzle()
                            self.view.model.emit("learn_success")
                            self.view.showNext()
                        else:
                            self.view.infobar.back_to_mainline()
                            self.view.showNext()
                else:
                    self.view.showNext()

            elif board.board.next.children:
                if self.view.model.lesson_game:
                    self.play_sound(move, board)
                    self.view.infobar.retry()

                # try to find this move in variations
                for i, vari in enumerate(board.board.next.children):
                    for node in vari:
                        if (
                            not isinstance(node, str)
                            and node.lastMove == move.move
                            and node.plyCount == board.ply + 1
                        ):
                            # replay variation move
                            self.view.setShownBoard(node.pieceBoard)
                            return

                # create new variation
                new_vari = self.view.model.add_variation(board, [move])
                self.view.setShownBoard(new_vari[-1])

            else:
                if self.view.model.lesson_game:
                    self.play_sound(move, board)
                    self.view.infobar.retry()

                # create new variation
                new_vari = self.view.model.add_variation(board, [move])
                self.view.setShownBoard(new_vari[-1])

    def emit_move_signal(self, cord0, cord1, promotion=None):
        # Game end can change cord0 to None while dragging a piece
        if cord0 is None:
            return
        gating = None
        board = self.getBoard()
        color = board.color
        # Ask player for which piece to promote into. If this move does not
        # include a promotion, QUEEN will be sent as a dummy value, but not used
        if (
            promotion is None
            and board[cord0].sign == PAWN
            and cord1.cord in board.PROMOTION_ZONE[color]
            and self.variant.variant != SITTUYINCHESS
        ):
            if len(self.variant.PROMOTIONS) == 1:
                promotion = lmove.PROMOTE_PIECE(self.variant.PROMOTIONS[0])
            elif self.variant.variant == LIGHTBRIGADECHESS:
                promotion = lmove.PROMOTE_PIECE(
                    QUEEN_PROMOTION if color == WHITE else KNIGHT_PROMOTION
                )
            else:
                if conf.get("autoPromote"):
                    promotion = lmove.PROMOTE_PIECE(QUEEN_PROMOTION)
                else:
                    promotion = self.getPromotion()
                    if promotion is None:
                        # Put back pawn moved be d'n'd
                        self.view.runAnimation(redraw_misc=False)
                        return
        if (
            promotion is None
            and board[cord0].sign == PAWN
            and cord0.cord in board.PROMOTION_ZONE[color]
            and self.variant.variant == SITTUYINCHESS
        ):
            # no promotion allowed if we have queen
            if board.board.boards[color][QUEEN]:
                promotion = None
            # in place promotion
            elif cord1.cord in board.PROMOTION_ZONE[color]:
                promotion = lmove.PROMOTE_PIECE(QUEEN_PROMOTION)
            # queen move promotion (but not a pawn capture!)
            elif board[cord1] is None and (cord0.cord + cord1.cord) % 2 == 1:
                promotion = lmove.PROMOTE_PIECE(QUEEN_PROMOTION)

        holding = board.board.holding[color]
        if self.variant.variant == SCHESS:
            moved = board[cord0].sign
            hawk = holding[HAWK] > 0
            elephant = holding[ELEPHANT] > 0
            if (hawk or elephant) and cord0.cord in iterBits(board.board.virgin[color]):
                castling = moved == KING and abs(cord0.x - cord1.x) == 2
                gating = self.getGating(castling, hawk, elephant)

        if gating is not None:
            if gating in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK):
                side = 0 if cord0.x - cord1.x == 2 else 1
                rcord = board.board.ini_rooks[color][side]
                move = Move(lmovegen.newMove(rcord, cord0.cord, gating))
            else:
                move = Move(lmovegen.newMove(cord0.cord, cord1.cord, gating))
        elif cord0.x < 0 or cord0.x > self.FILES - 1:
            move = Move(lmovegen.newMove(board[cord0].piece, cord1.cord, DROP))
        else:
            move = Move(cord0, cord1, board, promotion)

        if (
            (self.view.model.curplayer.__type__ == LOCAL or self.view.model.examined)
            and self.view.shownIsMainLine()
            and self.view.model.boards[-1] == board
            and self.view.model.status == RUNNING
        ):
            # emit move
            if self.setup_position:
                self.emit("piece_moved", (cord0, cord1), board[cord0].color)
            else:
                self.emit("piece_moved", move, color)
                if self.view.model.examined:
                    self.view.model.connection.bm.sendMove(toAN(board, move))
        else:
            self.play_or_add_move(board, move)

    def actionActivate(self, widget, key):
        """Put actions from a menu or similar"""
        curplayer = self.view.model.curplayer
        if key == "call_flag":
            self.emit("action", FLAG_CALL, curplayer, None)
        elif key == "abort":
            self.emit("action", ABORT_OFFER, curplayer, None)
        elif key == "adjourn":
            self.emit("action", ADJOURN_OFFER, curplayer, None)
        elif key == "draw":
            self.emit("action", DRAW_OFFER, curplayer, None)
        elif key == "resign":
            self.emit("action", RESIGNATION, curplayer, None)
        elif key == "ask_to_move":
            self.emit("action", HURRY_ACTION, curplayer, None)
        elif key == "undo1":
            board = self.view.model.getBoardAtPly(
                self.view.shown, variation=self.view.shown_variation_idx
            )
            if board.board.next is not None or board.board.children:
                return
            if not self.view.shownIsMainLine():
                self.view.model.undo_in_variation(board)
                return

            waitingplayer = self.view.model.waitingplayer
            if (
                curplayer.__type__ == LOCAL
                and (
                    waitingplayer.__type__ == ARTIFICIAL
                    or self.view.model.isPlayingICSGame()
                )
                and self.view.model.ply - self.view.model.lowply > 1
            ):
                self.emit("action", TAKEBACK_OFFER, curplayer, 2)
            else:
                self.emit("action", TAKEBACK_OFFER, curplayer, 1)
        elif key == "pause1":
            self.emit("action", PAUSE_OFFER, curplayer, None)
        elif key == "resume1":
            self.emit("action", RESUME_OFFER, curplayer, None)

    def shownChanged(self, view, shown):
        if self.view is None:
            return
        self.lockedPly = self.view.shown
        self.possibleBoards[self.lockedPly] = self._genPossibleBoards(self.lockedPly)
        if self.view.shown - 2 in self.possibleBoards:
            del self.possibleBoards[self.view.shown - 2]

    def moves_undone(self, gamemodel, moves):
        self.view.selected = None
        self.view.active = None
        self.view.hover = None
        self.view.dragged_piece = None
        self.view.setPremove(None, None, None, None)
        if not self.view.model.examined:
            self.currentState = self.lockedNormalState

    def game_ended(self, gamemodel, reason):
        self.selected_last = None
        self.view.selected = None
        self.view.active = None
        self.view.hover = None
        self.view.dragged_piece = None
        self.view.setPremove(None, None, None, None)
        self.currentState = self.normalState

        self.view.startAnimation()

    def game_started(self, gamemodel):
        if self.view.model.lesson_game:
            if "FEN" in gamemodel.tags:
                if gamemodel.orientation != gamemodel.starting_color:
                    self.view.showNext()
            else:
                self.view.infobar.get_next_puzzle()
                self.view.model.emit("learn_success")

    def getBoard(self):
        return self.view.model.getBoardAtPly(
            self.view.shown, self.view.shown_variation_idx
        )

    def isLastPlayed(self, board):
        return board == self.view.model.boards[-1]

    def setLocked(self, locked):
        do_animation = False

        if (
            locked
            and self.isLastPlayed(self.getBoard())
            and self.view.model.status == RUNNING
        ):
            if self.view.model.status != RUNNING:
                self.view.selected = None
                self.view.active = None
                self.view.hover = None
                self.view.dragged_piece = None
                do_animation = True

            if self.currentState == self.selectedState:
                self.currentState = self.lockedSelectedState
            elif self.currentState == self.activeState:
                self.currentState = self.lockedActiveState
            else:
                self.currentState = self.lockedNormalState
        else:
            if self.currentState == self.lockedSelectedState:
                self.currentState = self.selectedState
            elif self.currentState == self.lockedActiveState:
                self.currentState = self.activeState
            else:
                self.currentState = self.normalState

        if do_animation:
            self.view.startAnimation()

    def setStateSelected(self):
        if self.currentState in (
            self.lockedNormalState,
            self.lockedSelectedState,
            self.lockedActiveState,
        ):
            self.currentState = self.lockedSelectedState
        else:
            self.view.setPremove(None, None, None, None)
            self.currentState = self.selectedState

    def setStateActive(self):
        if self.currentState in (
            self.lockedNormalState,
            self.lockedSelectedState,
            self.lockedActiveState,
        ):
            self.currentState = self.lockedActiveState
        else:
            self.view.setPremove(None, None, None, None)
            self.currentState = self.activeState

    def setStateNormal(self):
        if self.currentState in (
            self.lockedNormalState,
            self.lockedSelectedState,
            self.lockedActiveState,
        ):
            self.currentState = self.lockedNormalState
        else:
            self.view.setPremove(None, None, None, None)
            self.currentState = self.normalState

    def color(self, event):
        state = event.get_state()
        if (
            state & Gdk.ModifierType.SHIFT_MASK
            and state & Gdk.ModifierType.CONTROL_MASK
        ):
            return "Y"
        elif state & Gdk.ModifierType.SHIFT_MASK:
            return "R"
        elif state & Gdk.ModifierType.CONTROL_MASK:
            return "B"
        else:
            return "G"

    def button_press(self, widget, event):
        if event.button == 3:
            # first we will draw a circle
            cord = self.currentState.point2Cord(event.x, event.y, self.color(event))
            if (
                cord is None
                or cord.x < 0
                or cord.x > self.FILES
                or cord.y < 0
                or cord.y > self.RANKS
            ):
                return
            self.pre_arrow_from = cord
            self.view.pre_circle = cord
            self.view.redrawCanvas()
            return
        else:
            # remove all circles and arrows
            need_redraw = False
            if self.view.arrows:
                self.view.arrows.clear()
                need_redraw = True
            if self.view.circles:
                self.view.circles.clear()
                need_redraw = True
            if self.view.pre_arrow is not None:
                self.view.pre_arrow = None
                need_redraw = True
            if self.view.pre_circle is not None:
                self.view.pre_circle = None
                need_redraw = True
            if need_redraw:
                self.view.redrawCanvas()

        if self.game_preview:
            return
        return self.currentState.press(event.x, event.y, event.button)

    def button_release(self, widget, event):
        if event.button == 3:
            # remove or finalize circle/arrow as needed
            cord = self.currentState.point2Cord(event.x, event.y, self.color(event))
            if (
                cord is None
                or cord.x < 0
                or cord.x > self.FILES
                or cord.y < 0
                or cord.y > self.RANKS
            ):
                return
            if self.view.pre_circle == cord:
                if cord in self.view.circles:
                    self.view.circles.remove(cord)
                else:
                    self.view.circles.add(cord)
                self.view.pre_circle = None
                self.emit("shapes_changed")

            if self.view.pre_arrow is not None:
                if self.view.pre_arrow in self.view.arrows:
                    self.view.arrows.remove(self.view.pre_arrow)
                else:
                    self.view.arrows.add(self.view.pre_arrow)
                self.view.pre_arrow = None
                self.emit("shapes_changed")

            self.pre_arrow_from = None
            self.pre_arrow_to = None
            self.view.redrawCanvas()
            return

        if self.game_preview:
            return
        return self.currentState.release(event.x, event.y)

    def motion_notify(self, widget, event):
        to = self.currentState.point2Cord(event.x, event.y)
        if to is None or to.x < 0 or to.x > self.FILES or to.y < 0 or to.y > self.RANKS:
            return
        if self.pre_arrow_from is not None:
            if to != self.pre_arrow_from:
                # this will be an arrow
                if self.pre_arrow_to is not None and to != self.pre_arrow_to:
                    # first remove the old one
                    self.view.pre_arrow = None
                    self.view.redrawCanvas()

                arrow = self.pre_arrow_from, to
                if arrow != self.view.pre_arrow:
                    # draw the new arrow
                    self.view.pre_arrow = arrow
                    self.view.pre_circle = None
                    self.view.redrawCanvas()
                    self.pre_arrow_to = to

            elif self.view.pre_circle is None:
                # back to circle
                self.view.pre_arrow = None
                self.view.pre_circle = to
                self.view.redrawCanvas()

        return self.currentState.motion(event.x, event.y)

    def leave_notify(self, widget, event):
        return self.currentState.leave(event.x, event.y)

    def key_pressed(self, keyname):
        if keyname in "PNBRQKMFSOox12345678abcdefgh":
            self.keybuffer += keyname

        elif keyname == "minus":
            self.keybuffer += "-"

        elif keyname == "at":
            self.keybuffer += "@"

        elif keyname == "equal":
            self.keybuffer += "="

        elif keyname == "Return" and self.keybuffer != "":
            color = self.view.model.boards[-1].color
            board = self.view.model.getBoardAtPly(
                self.view.shown, self.view.shown_variation_idx
            )
            try:
                move = parseAny(board, self.keybuffer)
            except ParsingError:
                self.keybuffer = ""
                return

            if validate(board, move):
                if (
                    (
                        self.view.model.curplayer.__type__ == LOCAL
                        or self.view.model.examined
                    )
                    and self.view.shownIsMainLine()
                    and self.view.model.boards[-1] == board
                    and self.view.model.status == RUNNING
                ):
                    # emit move
                    self.emit("piece_moved", move, color)
                    if self.view.model.examined:
                        self.view.model.connection.bm.sendMove(toAN(board, move))
                else:
                    self.play_or_add_move(board, move)
            self.keybuffer = ""

        elif keyname == "BackSpace":
            self.keybuffer = self.keybuffer[:-1] if self.keybuffer else ""

    def _genPossibleBoards(self, ply):
        possible_boards = []
        if self.setup_position:
            return possible_boards
        if len(self.view.model.players) == 2 and self.view.model.isEngine2EngineGame():
            return possible_boards
        curboard = self.view.model.getBoardAtPly(ply, self.view.shown_variation_idx)
        for lmove_item in lmovegen.genAllMoves(curboard.board.clone()):
            move = Move(lmove_item)
            board = curboard.move(move)
            possible_boards.append(board)
        return possible_boards


class BoardState:
    """
    There are 6 total BoardStates:
    NormalState, ActiveState, SelectedState
    LockedNormalState, LockedActiveState, LockedSelectedState

    The board state is Locked while it is the opponents turn.
    The board state is not Locked during your turn.
    (Locked states are not used when BoardControl setup_position is True.)

    Normal/Locked State - No pieces or cords are selected
    Active State - A piece is currently being dragged by the mouse
    Selected State - A cord is currently selected
    """

    def __init__(self, board):
        self.parent = board
        self.view = board.view
        self.lastMotionCord = None

        self.RANKS = self.view.model.boards[0].RANKS
        self.FILES = self.view.model.boards[0].FILES

    def getBoard(self):
        return self.view.model.getBoardAtPly(
            self.view.shown, self.view.shown_variation_idx
        )

    def validate(self, cord0, cord1):
        if cord0 is None or cord1 is None:
            return False
        # prevent accidental null move creation
        if cord0 == cord1 and self.parent.variant.variant != SITTUYINCHESS:
            return False
        if self.getBoard()[cord0] is None:
            return False

        if self.parent.setup_position:
            # prevent moving pieces inside holding
            if (cord0.x < 0 or cord0.x > self.FILES - 1) and (
                cord1.x < 0 or cord1.x > self.FILES - 1
            ):
                return False
            else:
                return True

        if cord1.x < 0 or cord1.x > self.FILES - 1:
            return False
        if cord0.x < 0 or cord0.x > self.FILES - 1:
            # drop
            return validate(
                self.getBoard(),
                Move(lmovegen.newMove(self.getBoard()[cord0].piece, cord1.cord, DROP)),
            )
        else:
            return validate(self.getBoard(), Move(cord0, cord1, self.getBoard()))

    def transPoint(self, x_loc, y_loc):
        xc_loc, yc_loc, side = (
            self.view.square[0],
            self.view.square[1],
            self.view.square[3],
        )
        x_loc, y_loc = self.view.invmatrix.transform_point(x_loc, y_loc)
        y_loc -= yc_loc
        x_loc -= xc_loc

        y_loc /= float(side)
        x_loc /= float(side)
        return x_loc, self.RANKS - y_loc

    def point2Cord(self, x_loc, y_loc, color=None):
        point = self.transPoint(x_loc, y_loc)
        p0_loc, p1_loc = point[0], point[1]
        if self.parent.variant.variant in DROP_VARIANTS:
            if (
                not -3 <= int(p0_loc) <= self.FILES + 2
                or not 0 <= int(p1_loc) <= self.RANKS - 1
            ):
                return None
        else:
            if (
                not 0 <= int(p0_loc) <= self.FILES - 1
                or not 0 <= int(p1_loc) <= self.RANKS - 1
            ):
                return None
        return Cord(int(p0_loc) if p0_loc >= 0 else int(p0_loc) - 1, int(p1_loc), color)

    def isSelectable(self, cord):
        # Simple isSelectable method, disabling selecting cords out of bound etc
        if not cord:
            return False
        if self.parent.setup_position:
            return True
        if self.parent.variant.variant in DROP_VARIANTS:
            if (not -3 <= cord.x <= self.FILES + 2) or (
                not 0 <= cord.y <= self.RANKS - 1
            ):
                return False
        else:
            if (not 0 <= cord.x <= self.FILES - 1) or (
                not 0 <= cord.y <= self.RANKS - 1
            ):
                return False
        return True

    def press(self, x_loc, y_loc, button):
        pass

    def release(self, x_loc, y_loc):
        pass

    def motion(self, x_loc, y_loc):
        cord = self.point2Cord(x_loc, y_loc)
        if self.lastMotionCord == cord:
            return
        self.lastMotionCord = cord
        if cord and self.isSelectable(cord):
            if not self.view.model.isPlayingICSGame():
                self.view.hover = cord
        else:
            self.view.hover = None

    def leave(self, x_loc, y_loc):
        allocation = self.parent.get_allocation()
        if not (0 <= x_loc < allocation.width and 0 <= y_loc < allocation.height):
            self.view.hover = None


class LockedBoardState(BoardState):
    """
    Parent of LockedNormalState, LockedActiveState, LockedSelectedState

    The board is in one of the three Locked states during the opponent's turn.
    """

    def __init__(self, board):
        BoardState.__init__(self, board)

    def isAPotentiallyLegalNextMove(self, cord0, cord1):
        """Determines whether the given move is at all legally possible
        as the next move after the player who's turn it is makes their move
        Note: This doesn't always return the correct value, such as when
        BoardControl.setLocked() has been called and we've begun a drag,
        but view.shown and BoardControl.lockedPly haven't been updated yet"""
        if cord0 is None or cord1 is None:
            return False
        if self.parent.lockedPly not in self.parent.possibleBoards:
            return False
        for board in self.parent.possibleBoards[self.parent.lockedPly]:
            if not board[cord0]:
                return False
            if validate(board, Move(cord0, cord1, board)):
                return True
        return False


class NormalState(BoardState):
    """
    It is the human player's turn and no pieces or cords are selected.
    """

    def isSelectable(self, cord):
        if not BoardState.isSelectable(self, cord):
            return False
        if self.parent.setup_position:
            return True
        try:
            board = self.getBoard()
            if board[cord] is None:
                return False  # We don't want empty cords
            elif board[cord].color != board.color:
                return False  # We shouldn't be able to select an opponent piece
        except IndexError:
            return False
        return True

    def press(self, x_loc, y_loc, button):
        self.parent.grab_focus()
        cord = self.point2Cord(x_loc, y_loc)
        if self.isSelectable(cord):
            self.view.dragged_piece = self.getBoard()[cord]
            self.view.active = cord
            self.parent.setStateActive()


class ActiveState(BoardState):
    """
    It is the human player's turn and a piece is being dragged by the mouse.
    """

    def isSelectable(self, cord):
        if not BoardState.isSelectable(self, cord):
            return False
        if self.parent.setup_position:
            return True
        return self.validate(self.view.active, cord)

    def release(self, x_loc, y_loc):
        cord = self.point2Cord(x_loc, y_loc)
        if (
            self.view.selected
            and cord != self.view.active
            and not self.validate(self.view.selected, cord)
        ):
            if not self.parent.setup_position:
                preferencesDialog.SoundTab.playAction("invalidMove")
        if not cord:
            self.view.active = None
            self.view.selected = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateNormal()

        # When in the mixed active/selected state
        elif self.view.selected:
            # Move when releasing on a good cord
            if self.validate(self.view.selected, cord):
                self.parent.setStateNormal()
                # It is important to emit_move_signal after setting state
                # as listeners of the function probably will lock the board
                self.view.dragged_piece = None
                self.parent.emit_move_signal(self.view.selected, cord)
                if self.parent.setup_position:
                    if not (
                        self.view.selected.x < 0
                        or self.view.selected.x > self.FILES - 1
                    ):
                        self.view.selected = None
                    else:
                        # enable stamping with selected holding pieces
                        self.parent.setStateSelected()
                else:
                    self.view.selected = None
                self.view.active = None
            elif (
                cord
                == self.view.active
                == self.view.selected
                == self.parent.selected_last
            ):
                # user clicked (press+release) same piece twice, so unselect it
                self.view.active = None
                self.view.selected = None
                self.view.dragged_piece = None
                self.view.startAnimation()
                self.parent.setStateNormal()
                if self.parent.variant.variant == SITTUYINCHESS:
                    self.parent.emit_move_signal(self.view.selected, cord)
            else:  # leave last selected piece selected
                self.view.active = None
                self.view.dragged_piece = None
                self.view.startAnimation()
                self.parent.setStateSelected()

        # If dragged and released on a possible cord
        elif self.validate(self.view.active, cord):
            self.parent.setStateNormal()
            self.view.dragged_piece = None
            # removig piece from board
            if self.parent.setup_position and (cord.x < 0 or cord.x > self.FILES - 1):
                self.view.startAnimation()
            self.parent.emit_move_signal(self.view.active, cord)
            self.view.active = None

        # Select last piece user tried to move or that was selected
        elif self.view.active or self.view.selected:
            self.view.selected = (
                self.view.active if self.view.active else self.view.selected
            )
            self.view.active = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateSelected()

        # Send back, if dragging to a not possible cord
        else:
            self.view.active = None
            # Send the piece back to its original cord
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateNormal()

        self.parent.selected_last = self.view.selected

    def motion(self, x_loc, y_loc):
        BoardState.motion(self, x_loc, y_loc)
        fcord = self.view.active
        if not fcord:
            return
        piece = self.getBoard()[fcord]
        if not piece:
            return
        elif piece.color != self.getBoard().color:
            if not self.parent.setup_position:
                return

        side = self.view.square[3]
        co_loc, si_loc = self.view.matrix[0], self.view.matrix[1]
        point = self.transPoint(
            x_loc - side * (co_loc + si_loc) / 2.0,
            y_loc + side * (co_loc - si_loc) / 2.0,
        )
        if not point:
            return
        x_loc, y_loc = point

        if piece.x != x_loc or piece.y != y_loc:
            if piece.x:
                paintbox = self.view.cord2RectRelative(piece.x, piece.y)
            else:
                paintbox = self.view.cord2RectRelative(self.view.active)
            paintbox = join(paintbox, self.view.cord2RectRelative(x_loc, y_loc))
            piece.x = x_loc
            piece.y = y_loc
            self.view.redrawCanvas(rect(paintbox))


class SelectedState(BoardState):
    """
    It is the human player's turn and a cord is selected.
    """

    def isSelectable(self, cord):
        if not BoardState.isSelectable(self, cord):
            return False
        if self.parent.setup_position:
            return True
        try:
            board = self.getBoard()
            if board[cord] is not None and board[cord].color == board.color:
                return True  # Select another piece
        except IndexError:
            return False
        return self.validate(self.view.selected, cord)

    def press(self, x_loc, y_loc, button):
        cord = self.point2Cord(x_loc, y_loc)
        # Unselecting by pressing the selected cord, or marking the cord to be
        # moved to. We don't unset self.view.selected, so ActiveState can handle
        # things correctly
        if self.isSelectable(cord):
            if self.parent.setup_position:
                color_ok = True
            else:
                color_ok = (
                    self.getBoard()[cord] is not None
                    and self.getBoard()[cord].color == self.getBoard().color
                )
            if (
                self.view.selected
                and self.view.selected != cord
                and color_ok
                and not self.validate(self.view.selected, cord)
            ):
                # corner case encountered:
                # user clicked (press+release) a piece, then clicked (no release yet)
                # a different piece and dragged it somewhere else. Since
                # ActiveState.release() will use self.view.selected as the source piece
                # rather than self.view.active, we need to update it here
                self.view.selected = cord  # re-select new cord

            self.view.dragged_piece = self.getBoard()[cord]
            self.view.active = cord
            self.parent.setStateActive()

        else:  # Unselecting by pressing an inactive cord
            self.view.selected = None
            self.parent.setStateNormal()
            if not self.parent.setup_position:
                preferencesDialog.SoundTab.playAction("invalidMove")


class LockedNormalState(LockedBoardState):
    """
    It is the opponent's turn and no piece or cord is selected.
    """

    def isSelectable(self, cord):
        if not BoardState.isSelectable(self, cord):
            return False
        if not self.parent.allowPremove:
            return False  # Don't allow premove if neither player is human
        try:
            board = self.getBoard()
            if board[cord] is None:
                return False  # We don't want empty cords
            elif board[cord].color == board.color:
                return False  # We shouldn't be able to select an opponent piece
        except IndexError:
            return False
        return True

    def press(self, x, y, button):
        self.parent.grab_focus()
        cord = self.point2Cord(x, y)
        if self.isSelectable(cord):
            self.view.dragged_piece = self.getBoard()[cord]
            self.view.active = cord
            self.parent.setStateActive()

        # reset premove if mouse right-clicks or clicks one of the premove cords
        if button == 3:  # right-click
            self.view.setPremove(None, None, None, None)
            self.view.startAnimation()
        elif cord == self.view.premove0 or cord == self.view.premove1:
            self.view.setPremove(None, None, None, None)
            self.view.startAnimation()


class LockedActiveState(LockedBoardState):
    """
    It is the opponent's turn and a piece is being dragged by the mouse.
    """

    def isSelectable(self, cord):
        if not BoardState.isSelectable(self, cord):
            return False
        return self.isAPotentiallyLegalNextMove(self.view.active, cord)

    def release(self, x_loc, y_loc):
        cord = self.point2Cord(x_loc, y_loc)
        if cord == self.view.active == self.view.selected == self.parent.selected_last:
            # User clicked (press+release) same piece twice, so unselect it
            self.view.active = None
            self.view.selected = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateNormal()
        elif (
            self.parent.allowPremove
            and self.view.selected
            and self.isAPotentiallyLegalNextMove(self.view.selected, cord)
        ):
            # In mixed locked selected/active state and user selects a valid premove cord
            board = self.getBoard()
            if (
                board[self.view.selected].sign == PAWN
                and cord.cord in board.PROMOTION_ZONE[1 - board.color]
            ):
                if len(board.PROMOTIONS) == 1:
                    promotion = lmove.PROMOTE_PIECE(board.PROMOTIONS[0])
                elif board.variant == LIGHTBRIGADECHESS:
                    promotion = lmove.PROMOTE_PIECE(
                        QUEEN_PROMOTION
                        if 1 - board.color == WHITE
                        else KNIGHT_PROMOTION
                    )
                else:
                    if conf.get("autoPromote"):
                        promotion = lmove.PROMOTE_PIECE(QUEEN_PROMOTION)
                    else:
                        promotion = self.parent.getPromotion()
            else:
                promotion = None
            self.view.setPremove(
                board[self.view.selected],
                self.view.selected,
                cord,
                self.view.shown + 2,
                promotion,
            )
            self.view.selected = None
            self.view.active = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateNormal()
        elif self.parent.allowPremove and self.isAPotentiallyLegalNextMove(
            self.view.active, cord
        ):
            # User drags a piece to a valid premove square
            board = self.getBoard()
            if (
                board[self.view.active].sign == PAWN
                and cord.cord in board.PROMOTION_ZONE[1 - board.color]
            ):
                if len(board.PROMOTIONS) == 1:
                    promotion = lmove.PROMOTE_PIECE(board.PROMOTIONS[0])
                elif board.variant == LIGHTBRIGADECHESS:
                    promotion = lmove.PROMOTE_PIECE(
                        QUEEN_PROMOTION
                        if 1 - board.color == WHITE
                        else KNIGHT_PROMOTION
                    )
                else:
                    if conf.get("autoPromote"):
                        promotion = lmove.PROMOTE_PIECE(QUEEN_PROMOTION)
                    else:
                        promotion = self.parent.getPromotion()
            else:
                promotion = None
            self.view.setPremove(
                self.getBoard()[self.view.active],
                self.view.active,
                cord,
                self.view.shown + 2,
                promotion,
            )
            self.view.selected = None
            self.view.active = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateNormal()
        elif self.view.active or self.view.selected:
            # Select last piece user tried to move or that was selected
            self.view.selected = (
                self.view.active if self.view.active else self.view.selected
            )
            self.view.active = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateSelected()
        else:
            self.view.active = None
            self.view.selected = None
            self.view.dragged_piece = None
            self.view.startAnimation()
            self.parent.setStateNormal()

        self.parent.selected_last = self.view.selected

    def motion(self, x_loc, y_loc):
        BoardState.motion(self, x_loc, y_loc)
        fcord = self.view.active
        if not fcord:
            return
        piece = self.getBoard()[fcord]
        if not piece or piece.color == self.getBoard().color:
            return

        side = self.view.square[3]
        co_loc, si_loc = self.view.matrix[0], self.view.matrix[1]
        point = self.transPoint(
            x_loc - side * (co_loc + si_loc) / 2.0,
            y_loc + side * (co_loc - si_loc) / 2.0,
        )
        if not point:
            return
        x_loc, y_loc = point

        if piece.x != x_loc or piece.y != y_loc:
            if piece.x:
                paintbox = self.view.cord2RectRelative(piece.x, piece.y)
            else:
                paintbox = self.view.cord2RectRelative(self.view.active)
            paintbox = join(paintbox, self.view.cord2RectRelative(x_loc, y_loc))
            piece.x = x_loc
            piece.y = y_loc
            self.view.redrawCanvas(rect(paintbox))


class LockedSelectedState(LockedBoardState):
    """
    It is the opponent's turn and a cord is selected.
    """

    def isSelectable(self, cord):
        if not BoardState.isSelectable(self, cord):
            return False
        try:
            board = self.getBoard()
            if board[cord] is not None and board[cord].color != board.color:
                return True  # Select another piece
        except IndexError:
            return False
        return self.isAPotentiallyLegalNextMove(self.view.selected, cord)

    def motion(self, x_loc, y_loc):
        cord = self.point2Cord(x_loc, y_loc)
        if self.lastMotionCord == cord:
            self.view.hover = cord
            return
        self.lastMotionCord = cord
        if cord and self.isAPotentiallyLegalNextMove(self.view.selected, cord):
            if not self.view.model.isPlayingICSGame():
                self.view.hover = cord
        else:
            self.view.hover = None

    def press(self, x_loc, y_loc, button):
        cord = self.point2Cord(x_loc, y_loc)
        # Unselecting by pressing the selected cord, or marking the cord to be
        # moved to. We don't unset self.view.selected, so ActiveState can handle
        # things correctly
        if self.isSelectable(cord):
            if (
                self.view.selected
                and self.view.selected != cord
                and self.getBoard()[cord] is not None
                and self.getBoard()[cord].color != self.getBoard().color
                and not self.isAPotentiallyLegalNextMove(self.view.selected, cord)
            ):
                # corner-case encountered (see comment in SelectedState.press)
                self.view.selected = cord  # re-select new cord

            self.view.dragged_piece = self.getBoard()[cord]
            self.view.active = cord
            self.parent.setStateActive()

        else:  # Unselecting by pressing an inactive cord
            self.view.selected = None
            self.parent.setStateNormal()

        # reset premove if mouse right-clicks or clicks one of the premove cords
        if button == 3:  # right-click
            self.view.setPremove(None, None, None, None)
            self.view.startAnimation()
        elif cord == self.view.premove0 or cord == self.view.premove1:
            self.view.setPremove(None, None, None, None)
            self.view.startAnimation()



"""
Internal datagram codes used in ICC level 2. unit format
see https://www.chessclub.com/user/resources/formats/formats.txt

This is somewhat similar to fics block mode. The main difference is fics
block mode works 'all or nothing' meaning when you switch to use it
all commands you send and resposes you get have to (or will) use it.
ICC datagram format are used for responses only and is more forgiving.
You can set it for individual responses one by one or for all at once
at login time.
"""

MY_ICC_PREFIX = "#"

UNIT_START = chr(25) + "["
UNIT_END = chr(25) + "]"
B_UNIT_END = b"\x19]"

DTGR_START = chr(25) + "("
DTGR_END = chr(25) + ")"
B_DTGR_END = b"\x19)"

DG_WHO_AM_I = 0
DG_PLAYER_ARRIVED = 1
DG_PLAYER_LEFT = 2
DG_BULLET = 3
DG_BLITZ = 4
DG_STANDARD = 5
DG_WILD = 6
DG_BUGHOUSE = 7
DG_TIMESTAMP = 8
DG_TITLES = 9
DG_OPEN = 10
DG_STATE = 11
DG_GAME_STARTED = 12
DG_GAME_RESULT = 13
DG_EXAMINED_GAME_IS_GONE = 14
DG_MY_GAME_STARTED = 15
DG_MY_GAME_RESULT = 16
DG_MY_GAME_ENDED = 17
DG_STARTED_OBSERVING = 18
DG_STOP_OBSERVING = 19
DG_PLAYERS_IN_MY_GAME = 20
DG_OFFERS_IN_MY_GAME = 21
DG_TAKEBACK = 22
DG_BACKWARD = 23
DG_SEND_MOVES = 24
DG_MOVE_LIST = 25
DG_KIBITZ = 26
DG_PEOPLE_IN_MY_CHANNEL = 27
DG_CHANNEL_TELL = 28
DG_MATCH = 29
DG_MATCH_REMOVED = 30
DG_PERSONAL_TELL = 31
DG_SHOUT = 32
DG_MOVE_ALGEBRAIC = 33
DG_MOVE_SMITH = 34
DG_MOVE_TIME = 35
DG_MOVE_CLOCK = 36
DG_BUGHOUSE_HOLDINGS = 37
DG_SET_CLOCK = 38
DG_FLIP = 39
DG_ISOLATED_BOARD = 40
DG_REFRESH = 41
DG_ILLEGAL_MOVE = 42
DG_MY_RELATION_TO_GAME = 43
DG_PARTNERSHIP = 44
DG_SEES_SHOUTS = 45
DG_CHANNELS_SHARED = 46
DG_MY_VARIABLE = 47
DG_MY_STRING_VARIABLE = 48
DG_JBOARD = 49
DG_SEEK = 50
DG_SEEK_REMOVED = 51
DG_MY_RATING = 52
DG_SOUND = 53
DG_PLAYER_ARRIVED_SIMPLE = 55
DG_MSEC = 56
DG_BUGHOUSE_PASS = 57
DG_IP = 58
DG_CIRCLE = 59
DG_ARROW = 60
DG_MORETIME = 61
DG_PERSONAL_TELL_ECHO = 62
DG_SUGGESTION = 63
DG_NOTIFY_ARRIVED = 64
DG_NOTIFY_LEFT = 65
DG_NOTIFY_OPEN = 66
DG_NOTIFY_STATE = 67
DG_MY_NOTIFY_LIST = 68
DG_LOGIN_FAILED = 69
DG_FEN = 70
DG_TOURNEY_MATCH = 71
DG_GAMELIST_BEGIN = 72
DG_GAMELIST_ITEM = 73
DG_IDLE = 74
DG_ACK_PING = 75
DG_RATING_TYPE_KEY = 76
DG_GAME_MESSAGE = 77
DG_UNACCENTED = 78
DG_STRINGLIST_BEGIN = 79
DG_STRINGLIST_ITEM = 80
DG_DUMMY_RESPONSE = 81
DG_CHANNEL_QTELL = 82
DG_PERSONAL_QTELL = 83
DG_SET_BOARD = 84
DG_MATCH_ASSESSMENT = 85
DG_LOG_PGN = 86
DG_NEW_MY_RATING = 87
DG_LOSERS = 88
DG_UNCIRCLE = 89
DG_UNARROW = 90
DG_WSUGGEST = 91
DG_TEMPORARY_PASSWORD = 93
DG_MESSAGELIST_BEGIN = 94
DG_MESSAGELIST_ITEM = 95
DG_LIST = 96
DG_SJI_AD = 97
DG_RETRACT = 99
DG_MY_GAME_CHANGE = 100
DG_POSITION_BEGIN = 101
DG_TOURNEY = 103
DG_REMOVE_TOURNEY = 104
DG_DIALOG_START = 105
DG_DIALOG_DATA = 106
DG_DIALOG_DEFAULT = 107
DG_DIALOG_END = 108
DG_DIALOG_RELEASE = 109
DG_POSITION_BEGIN2 = 110
DG_PAST_MOVE = 111
DG_PGN_TAG = 112
DG_IS_VARIATION = 113
DG_PASSWORD = 114
DG_WILD_KEY = 116
DG_SWITCH_SERVERS = 120
DG_CRAZYHOUSE = 121
DG_SET2 = 124
DG_FIVEMINUTE = 125
DG_ONEMINUTE = 126
DG_TRANSLATIONOKAY = 129
DG_UID = 131
DG_KNOWS_FISCHER_RANDOM = 132
DG_COMMAND = 136
DG_TOURNEY_GAME_STARTED = 137
DG_TOURNEY_GAME_ENDED = 138
DG_MY_TURN = 139
DG_CORRESPONDENCE_RATING = 140
DG_DISABLE_PREMOVE = 141
DG_PSTAT = 142
DG_BOARDINFO = 143
DG_MOVE_LAG = 144
DG_FIFTEENMINUTE = 145
DG_PHRASELIST_UPDATE = 146
DG_PHRASELIST_ITEM = 147
DG_MENU_SPEAK = 148
DG_THREEMINUTE = 149
DG_FORTYFIVEMINUTE = 150
DG_CHESS960 = 151

SCN_UNKNOWN = 0
SCN_ILLEGAL_MOVE = 1
SCN_MOVE = 2
SCN_EDIT_EXAMINED = 3
SCN_PING_RESPONSE = 10
SCN_WEIRD = 11
SCN_REALLY_LOG_IN = 12
SCN_MOED = 13
SCN_EVENTS = 14
SCN_NEWS = 15
SCN_LOGOUT = 16
SCN_TIMEOUT = 17
SCN_CONNECT = 18
SCN_REALLY_QUIT = 19
SCN_LOGIN = 20
SCN_PASSWORD = 21
SCN_REGISTRATION = 22
SCN_EXTENSION = 23
SCN_AUTHENTICATION = 24
SCN_BAD = 25
SCN_AUTOMATIC = 26
SCN_CONFIRM = 27
SCN_MULTI_DISCARD = 28
SCN_IDLE = 29
SCN_ACK_PING = 30
SCN_MISCELLANEOUS = 31

CN_TELL = 101
CN_I = 102
CN_SHOUT = 103
CN_SHOUT0 = 104
CN_SLASH = 105
CN_WHO = 106
CN_SET = 107
CN_FLAG = 108
CN_SAY = 109
CN_CHANNELTELL = 110
CN_SSHOUT = 111
CN_BAD = 112
CN_KIBITZ = 113
CN_WHISPER = 114
CN_EXAMINE = 115
CN_MEXAMINE = 116
CN_COPYGAME0 = 117
CN_COPYGAME = 118
CN_FORWARD = 119
CN_BACK = 120
CN_MATCH = 121
CN_MATCH0 = 122
CN_ACCEPT = 123
CN_HELP0 = 124
CN_HELP = 125
CN_MORE = 126
CN_NEWS = 127
CN_NEWS0 = 128
CN_HISTORY = 129
CN_FINGER = 130
CN_VARS = 131
CN_UPSTATISTICS = 132
CN_UNEXAMINE = 133
CN_ADJOURN = 134
CN_ASSESS = 135
CN_OBSERVE0 = 136
CN_OBSERVE = 137
CN_FOLLOW0 = 138
CN_FOLLOW = 139
CN_ECO = 140
CN_STYLE = 141
CN_BELL = 142
CN_OPEN = 143
CN_DECLINE = 144
CN_REFRESH = 145
CN_RESIGNADJ = 146
CN_REVERT = 147
CN_RATED = 148
CN_RANK = 149
CN_MOVES = 150
CN_MAILOLDMOVES = 151
CN_MAILSTORED = 152
CN_MAILHELP = 153
CN_PENDING = 154
CN_GAMES = 155
CN_ABORT = 156
CN_ALLOBSERVERS = 157
CN_INCHANNEL = 158
CN_INFO = 159
CN_MORETIME = 160
CN_BEST = 161
CN_QUIT = 162
CN_QUOTA = 163
CN_LLOGONS = 164
CN_LHISTORY = 165
CN_LOGONS = 166
CN_TIME = 167
CN_TAKEBACK = 168
CN_SEARCH = 169
CN_SEARCH0 = 170
CN_SMOVES = 171
CN_SPOSITION = 172
CN_PASSWORD = 173
CN_MESSAGE = 174
CN_MESSAGE0 = 175
CN_CLEARMESSAGES = 176
CN_DATE = 177
CN_LIST = 178
CN_PLUS = 179
CN_MINUS = 180
CN_ZNOTL = 181
CN_FLIP = 182
CN_PROMOTE = 183
CN_EXPUNGE = 184
CN_IWCMATCH = 185
CN_LIMITS = 186
CN_PING = 187
CN_EXTEND = 188
CN_QTELL = 189
CN_GETPI = 190
CN_STARTSIMUL = 191
CN_GOTO = 192
CN_SETCLOCK = 193
CN_LIBLIST = 194
CN_LIBSAVE = 195
CN_LIBDELETE = 196
CN_LIBANNOTATE = 197
CN_LIBKEEPEXAM = 198
CN_PARTNER = 199
CN_PARTNER0 = 200
CN_PTELL = 201
CN_BUGWHO = 202
CN_WILDRANK = 204
CN_XOBSERVE = 205
CN_PRIMARY = 206
CN_DRAW = 207
CN_RESIGN = 208
CN_STATISTICS = 209
CN_STORED = 210
CN_CHANNELQTELL = 211
CN_XPARTNER = 212
CN_YFINGER = 213
CN_SEEKING = 214
CN_SOUGHT = 215
CN_SET2 = 216
CN_PLAY = 217
CN_UNSEEKING = 218
CN_AWAY = 219
CN_LAGSTATS = 220
CN_COMMANDS = 221
CN_REMATCH = 222
CN_REGISTER = 223
CN_RESUME = 224
CN_CIRCLE = 225
CN_ARROW = 226
CN_BLANKING = 227
CN_RELAY = 228
CN_LOADGAME = 229
CN_DRAWADJ = 230
CN_ABORTADJ = 231
CN_MAILNEWS = 232
CN_QSET = 233
CN_CC_START = 234
CN_CC_LIST = 235
CN_CC_MOVE = 236
CN_CC_DELETE = 237
CN_CC_QSTART = 238
CN_CC_QLIST = 239
CN_CC_QLABEL = 240
CN_CC_QDELETE = 241
CN_CC_QADJUDICATE = 242
CN_CC_ASK_DIRECTOR = 243
CN_LOADFEN = 244
CN_GETPX = 245
CN_UNRELAYED = 246
CN_NORELAY = 247
CN_REFER = 248
CN_PGN = 249
CN_SPGN = 250
CN_QFOLLOW = 251
CN_QUNFOLLOW = 252
CN_QMATCH = 253
CN_QPARTNER = 254
CN_ISREGNAME = 255
CN_REQUIRETICKET = 256
CN_ANNOTATE = 257
CN_CLEARBOARD = 258
CN_REQUEST_WIN = 259
CN_REQUEST_DRAW = 260
CN_REQUEST_ABORT = 261
CN_LOGPGN = 262
CN_RESULT = 263
CN_FEN = 264
CN_SFEN = 265
CN_SETGAMEPARAM = 266
CN_TAG = 267
CN_TOMOVE = 268
CN_REGENTRY = 269
CN_PERSONALINFO = 270
CN_EVENTS = 271
CN_QADDEVENT = 272
CN_GLISTGROUPS = 273
CN_GLISTMEMBERS = 274
CN_GINVITE = 275
CN_GJOIN = 276
CN_GLISTINVITED = 277
CN_GLISTJOINING = 278
CN_GDESCRIBE = 279
CN_GKICK = 280
CN_GBEST = 281
CN_SIMULIZE = 282
CN_GAMEID = 283
CN_THREEMIN = 92
CN_FIVEMINUTE = 284
CN_QIMPART = 285
CN_GMESSAGE = 286
CN_COMPLAIN = 287
CN_LASTTELLS = 288
CN_VIEW = 289
CN_SHOWADMIN = 290
CN_PSTAT = 291
CN_BOARDINFO = 292
